from os.path import getmtime
from contextlib import contextmanager
import re
import os
from pathlib import Path
import random
import uuid

import pytest
import numpy as np

import qcodes as qc
import qcodes.tests.dataset
from qcodes.dataset.experiment_container import Experiment,\
    load_experiment_by_name
from qcodes.dataset.data_set import (DataSet, load_by_guid, load_by_counter,
                                     load_by_id, load_by_run_spec,
                                     generate_dataset_table)
from qcodes.dataset.sqlite.database import get_db_version_and_newest_available_version
from qcodes.dataset.sqlite.connection import path_to_dbfile
from qcodes.dataset.database_extract_runs import extract_runs_into_db
from qcodes.dataset.sqlite.queries import get_experiments
from qcodes.tests.common import error_caused_by
from qcodes.dataset.measurements import Measurement
from qcodes import Station
from qcodes.tests.instrument_mocks import DummyInstrument
from qcodes.dataset.linked_datasets.links import Link


@contextmanager
def raise_if_file_changed(path_to_file: str):
    """
    Context manager that raises if a file is modified.
    On Windows, the OS modification time resolution is 100 ns
    """
    pre_operation_time = getmtime(path_to_file)
    # we don't want to catch and re-raise anything, since there is no clean-up
    # that we need to perform. Hence no try-except here
    yield
    post_operation_time = getmtime(path_to_file)
    if pre_operation_time != post_operation_time:
        raise RuntimeError(f'File {path_to_file} was modified.')


@pytest.fixture(scope='function')
def inst():
    """
    Dummy instrument for testing, ensuring that it's instance gets closed
    and removed from the global register of instruments, which, if not done,
    make break other tests
    """
    inst = DummyInstrument('extract_run_inst', gates=['back', 'plunger', 'cutter'])
    yield inst
    inst.close()


def test_missing_runs_raises(two_empty_temp_db_connections, some_interdeps):
    """
    Test that an error is raised if we attempt to extract a run not present in
    the source DB
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_exp_1 = Experiment(conn=source_conn)

    # make 5 runs in first experiment

    exp_1_run_ids = []
    for _ in range(5):

        source_dataset = DataSet(conn=source_conn, exp_id=source_exp_1.exp_id)
        exp_1_run_ids.append(source_dataset.run_id)
        source_dataset.set_interdependencies(some_interdeps[1])

        source_dataset.mark_started()

        for val in range(10):
            source_dataset.add_results([{name: val
                                         for name in some_interdeps[1].names}])
        source_dataset.mark_completed()

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    run_ids = [1, 8, 5, 3, 2, 4, 4, 4, 7, 8]
    wrong_ids = [8, 7, 8]

    expected_err = ("Error: not all run_ids exist in the source database. "
                    "The following run(s) is/are not present: "
                    f"{wrong_ids}")

    with pytest.raises(ValueError, match=re.escape(expected_err)):
        extract_runs_into_db(source_path, target_path, *run_ids)


def test_basic_extraction(two_empty_temp_db_connections, some_interdeps):
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    type_casters = {'numeric': float,
                    'array': (lambda x: np.array(x) if hasattr(x, '__iter__')
                              else np.array([x])),
                    'text': str}

    source_exp = Experiment(conn=source_conn)
    source_dataset = DataSet(conn=source_conn, name="basic_copy_paste_name")

    with pytest.raises(RuntimeError) as excinfo:
        extract_runs_into_db(source_path, target_path, source_dataset.run_id)

    assert error_caused_by(excinfo, ('Dataset not completed. An incomplete '
                                     'dataset can not be copied. The '
                                     'incomplete dataset has GUID: '
                                     f'{source_dataset.guid} and run_id: '
                                     f'{source_dataset.run_id}'))

    source_dataset.set_interdependencies(some_interdeps[0])

    source_dataset.parent_dataset_links = [Link(head=source_dataset.guid,
                                                tail=str(uuid.uuid4()),
                                                edge_type='test_link')]
    source_dataset.mark_started()

    for value in range(10):
        result = {ps.name: type_casters[ps.type](value)
                  for ps in some_interdeps[0].paramspecs}
        source_dataset.add_results([result])

    source_dataset.add_metadata('goodness', 'fair')
    source_dataset.add_metadata('test', True)


    source_dataset.mark_completed()

    assert source_dataset.run_id == source_dataset.captured_run_id

    extract_runs_into_db(source_path, target_path, source_dataset.run_id)

    target_exp = Experiment(conn=target_conn, exp_id=1)

    length1 = len(target_exp)
    assert length1 == 1

    # trying to insert the same run again should be a NOOP
    with raise_if_file_changed(target_path):
        extract_runs_into_db(source_path, target_path, source_dataset.run_id)

    assert len(target_exp) == length1

    target_dataset = DataSet(conn=target_conn, run_id=1)

    # Now make the interesting comparisons: are the target objects the same as
    # the source objects?

    assert source_dataset.the_same_dataset_as(target_dataset)

    source_data = source_dataset.get_parameter_data(*source_dataset.parameters.split(','))
    target_data = target_dataset.get_parameter_data(*target_dataset.parameters.split(','))

    for outkey, outval in source_data.items():
        for inkey, inval in outval.items():
            np.testing.assert_array_equal(inval, target_data[outkey][inkey])

    exp_attrs = ['name', 'sample_name', 'format_string', 'started_at',
                 'finished_at']

    for exp_attr in exp_attrs:
        assert getattr(source_exp, exp_attr) == getattr(target_exp, exp_attr)

    # trying to insert the same run again should be a NOOP
    with raise_if_file_changed(target_path):
        extract_runs_into_db(source_path, target_path, source_dataset.run_id)


def test_correct_experiment_routing(two_empty_temp_db_connections,
                                    some_interdeps):
    """
    Test that existing experiments are correctly identified AND that multiple
    insertions of the same runs don't matter (run insertion is idempotent)
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_exp_1 = Experiment(conn=source_conn)

    # make 5 runs in first experiment

    exp_1_run_ids = []
    for _ in range(5):

        source_dataset = DataSet(conn=source_conn, exp_id=source_exp_1.exp_id)
        exp_1_run_ids.append(source_dataset.run_id)

        source_dataset.set_interdependencies(some_interdeps[1])

        source_dataset.mark_started()

        for val in range(10):
            source_dataset.add_results([{name: val
                                         for name in some_interdeps[1].names}])
        source_dataset.mark_completed()

    # make a new experiment with 1 run

    source_exp_2 = Experiment(conn=source_conn)
    ds = DataSet(conn=source_conn, exp_id=source_exp_2.exp_id, name="lala")
    exp_2_run_ids = [ds.run_id]

    ds.set_interdependencies(some_interdeps[1])

    ds.mark_started()

    for val in range(10):
        ds.add_results([{name: val for name in some_interdeps[1].names}])

    ds.mark_completed()

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    # now copy 2 runs
    extract_runs_into_db(source_path, target_path, *exp_1_run_ids[:2])

    target_exp1 = Experiment(conn=target_conn, exp_id=1)

    assert len(target_exp1) == 2

    # copy two other runs, one of them already in
    extract_runs_into_db(source_path, target_path, *exp_1_run_ids[1:3])

    assert len(target_exp1) == 3

    # insert run from different experiment
    extract_runs_into_db(source_path, target_path, ds.run_id)

    assert len(target_exp1) == 3

    target_exp2 = Experiment(conn=target_conn, exp_id=2)

    assert len(target_exp2) == 1

    # finally insert every single run from experiment 1

    extract_runs_into_db(source_path, target_path, *exp_1_run_ids)

    # check for idempotency once more by inserting all the runs but in another
    # order
    with raise_if_file_changed(target_path):
        extract_runs_into_db(source_path, target_path, *exp_1_run_ids[::-1])

    target_exps = get_experiments(target_conn)

    assert len(target_exps) == 2
    assert len(target_exp1) == 5
    assert len(target_exp2) == 1

    # check that all the datasets match up
    for run_id in exp_1_run_ids + exp_2_run_ids:
        source_ds = DataSet(conn=source_conn, run_id=run_id)
        target_ds = load_by_guid(guid=source_ds.guid, conn=target_conn)

        assert source_ds.the_same_dataset_as(target_ds)

        source_data = source_ds.get_parameter_data(*source_ds.parameters.split(','))
        target_data = target_ds.get_parameter_data(*target_ds.parameters.split(','))

        for outkey, outval in source_data.items():
            for inkey, inval in outval.items():
                np.testing.assert_array_equal(inval, target_data[outkey][inkey])


def test_runs_from_different_experiments_raises(two_empty_temp_db_connections,
                                                some_interdeps):
    """
    Test that inserting runs from multiple experiments raises
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    source_exp_1 = Experiment(conn=source_conn)
    source_exp_2 = Experiment(conn=source_conn)

    # make 5 runs in first experiment

    exp_1_run_ids = []
    for _ in range(5):

        source_dataset = DataSet(conn=source_conn, exp_id=source_exp_1.exp_id)
        exp_1_run_ids.append(source_dataset.run_id)

        source_dataset.set_interdependencies(some_interdeps[1])

        source_dataset.mark_started()

        for val in range(10):
            source_dataset.add_results([{name: val
                                         for name in some_interdeps[1].names}])
        source_dataset.mark_completed()

    # make 5 runs in second experiment

    exp_2_run_ids = []
    for _ in range(5):

        source_dataset = DataSet(conn=source_conn, exp_id=source_exp_2.exp_id)
        exp_2_run_ids.append(source_dataset.run_id)

        source_dataset.set_interdependencies(some_interdeps[1])


        source_dataset.mark_started()

        for val in range(10):
            source_dataset.add_results([{name: val
                                         for name in some_interdeps[1].names}])
        source_dataset.mark_completed()

    run_ids = exp_1_run_ids + exp_2_run_ids
    source_exp_ids = np.unique([1, 2])
    matchstring = ('Did not receive runs from a single experiment\\. '
                   f'Got runs from experiments {source_exp_ids}')
    # make the matchstring safe to use as a regexp
    matchstring = matchstring.replace('[', '\\[').replace(']', '\\]')
    with pytest.raises(ValueError, match=matchstring):
        extract_runs_into_db(source_path, target_path, *run_ids)


def test_extracting_dataless_run(two_empty_temp_db_connections):
    """
    Although contrived, it could happen that a run with no data is extracted
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    Experiment(conn=source_conn)

    source_ds = DataSet(conn=source_conn)
    source_ds.mark_started()
    source_ds.mark_completed()

    extract_runs_into_db(source_path, target_path, source_ds.run_id)

    loaded_ds = DataSet(conn=target_conn, run_id=1)

    assert loaded_ds.the_same_dataset_as(source_ds)


def test_result_table_naming_and_run_id(two_empty_temp_db_connections,
                                        some_interdeps):
    """
    Check that a correct result table name is given and that a correct run_id
    is assigned
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    source_exp1 = Experiment(conn=source_conn)
    source_ds_1_1 = DataSet(conn=source_conn, exp_id=source_exp1.exp_id)
    source_ds_1_1.set_interdependencies(some_interdeps[1])

    source_ds_1_1.mark_started()
    source_ds_1_1.add_results([{name: 0.0
                                for name in some_interdeps[1].names}])
    source_ds_1_1.mark_completed()

    source_exp2 = Experiment(conn=source_conn)
    source_ds_2_1 = DataSet(conn=source_conn, exp_id=source_exp2.exp_id)
    source_ds_2_1.set_interdependencies(some_interdeps[1])
    source_ds_2_1.mark_started()
    source_ds_2_1.add_results([{name: 0.0
                                for name in some_interdeps[1].names}])
    source_ds_2_1.mark_completed()
    source_ds_2_2 = DataSet(conn=source_conn,
                            exp_id=source_exp2.exp_id,
                            name="customname")

    source_ds_2_2.set_interdependencies(some_interdeps[1])
    source_ds_2_2.mark_started()
    source_ds_2_2.add_results([{name: 0.0
                                for name in some_interdeps[1].names}])
    source_ds_2_2.mark_completed()

    extract_runs_into_db(source_path, target_path, source_ds_2_2.run_id)

    # The target ds ought to have a runs table "results-1-1"
    # and ought to be the same dataset as its "ancestor"
    target_ds = DataSet(conn=target_conn, run_id=1)

    assert target_ds.name == "customname"
    assert target_ds.table_name == "results-1-1"
    assert target_ds.the_same_dataset_as(source_ds_2_2)


def test_load_by_X_functions(two_empty_temp_db_connections,
                             some_interdeps):
    """
    Test some different loading functions
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    source_exp1 = Experiment(conn=source_conn)
    source_ds_1_1 = DataSet(conn=source_conn, exp_id=source_exp1.exp_id)

    source_exp2 = Experiment(conn=source_conn)
    source_ds_2_1 = DataSet(conn=source_conn, exp_id=source_exp2.exp_id)

    source_ds_2_2 = DataSet(conn=source_conn,
                            exp_id=source_exp2.exp_id,
                            name="customname")

    for ds in (source_ds_1_1, source_ds_2_1, source_ds_2_2):
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{name: 0.0 for name in some_interdeps[1].names}])
        ds.mark_completed()

    extract_runs_into_db(source_path, target_path, source_ds_2_2.run_id)
    extract_runs_into_db(source_path, target_path, source_ds_2_1.run_id)
    extract_runs_into_db(source_path, target_path, source_ds_1_1.run_id)

    test_ds = load_by_guid(source_ds_2_2.guid, target_conn)
    assert source_ds_2_2.the_same_dataset_as(test_ds)

    test_ds = load_by_id(1, target_conn)
    assert source_ds_2_2.the_same_dataset_as(test_ds)

    test_ds = load_by_run_spec(captured_run_id=source_ds_2_2.captured_run_id,
                               conn=target_conn)
    assert source_ds_2_2.the_same_dataset_as(test_ds)

    assert source_exp2.exp_id == 2

    # this is now the first run in the db so run_id is 1
    target_run_id = 1
    # and the experiment ids will be interchanged.
    target_exp_id = 1

    test_ds = load_by_counter(target_run_id, target_exp_id, target_conn)
    assert source_ds_2_2.the_same_dataset_as(test_ds)


def test_combine_runs(two_empty_temp_db_connections,
                      empty_temp_db_connection,
                      some_interdeps):
    """
    Test that datasets that are exported in random order from 2 datasets
    can be reloaded by the original captured_run_id and the experiment
    name.
    """
    source_conn_1, source_conn_2 = two_empty_temp_db_connections
    target_conn = empty_temp_db_connection

    source_1_exp = Experiment(conn=source_conn_1,
                              name='exp1',
                              sample_name='no_sample')
    source_1_datasets = [DataSet(conn=source_conn_1,
                                 exp_id=source_1_exp.exp_id) for i in range(10)]

    source_2_exp = Experiment(conn=source_conn_2,
                              name='exp2',
                              sample_name='no_sample')

    source_2_datasets = [DataSet(conn=source_conn_2,
                                 exp_id=source_2_exp.exp_id) for i in range(10)]

    source_all_datasets = source_1_datasets + source_2_datasets

    shuffled_datasets = source_all_datasets.copy()
    random.shuffle(shuffled_datasets)

    for ds in source_all_datasets:
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{name: 0.0 for name in some_interdeps[1].names}])
        ds.mark_completed()

    # now let's insert all datasets in random order
    for ds in shuffled_datasets:
        extract_runs_into_db(ds.conn.path_to_dbfile,
                             target_conn.path_to_dbfile, ds.run_id)

    for ds in source_all_datasets:
        loaded_ds = load_by_run_spec(captured_run_id=ds.captured_run_id,
                                     experiment_name=ds.exp_name,
                                     conn=target_conn)
        assert ds.the_same_dataset_as(loaded_ds)

    for ds in source_all_datasets:
        loaded_ds = load_by_run_spec(captured_run_id=ds.captured_counter,
                                     experiment_name=ds.exp_name,
                                     conn=target_conn)
        assert ds.the_same_dataset_as(loaded_ds)

    # Now test that we generate the correct table for the guids above
    # this could be split out into its own test
    # but the test above has the useful side effect of
    # setting up datasets for this test.
    guids = [ds.guid for ds in source_all_datasets]

    table = generate_dataset_table(guids, conn=target_conn)
    lines = table.split('\n')
    headers = re.split(r'\s+', lines[0].strip())

    cfg = qc.config
    guid_comp = cfg['GUID_components']

    # borrowed fallback logic from generate_guid
    sampleint = guid_comp['sample']
    if sampleint == 0:
        sampleint = int('a'*8, base=16)

    for i in range(2, len(lines)):
        split_line = re.split(r'\s+', lines[i].strip())
        mydict = {headers[j]: split_line[j] for j in range(len(split_line))}
        ds = load_by_guid(guids[i-2], conn=target_conn)
        assert ds.captured_run_id == int(mydict['captured_run_id'])
        assert ds.captured_counter == int(mydict['captured_counter'])
        assert ds.exp_name == mydict['experiment_name']
        assert ds.sample_name == mydict['sample_name']
        assert int(mydict['sample_id']) == sampleint
        assert guid_comp['location'] == int(mydict['location'])
        assert guid_comp['work_station'] == int(mydict['work_station'])


def test_copy_datasets_and_add_new(two_empty_temp_db_connections,
                                   some_interdeps):
    """
    Test that new runs get the correct captured_run_id and captured_counter
    when adding on top of a dataset with partial exports
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_exp_1 = Experiment(conn=source_conn,
                              name='exp1',
                              sample_name='no_sample')
    source_exp_2 = Experiment(conn=source_conn,
                              name='exp2',
                              sample_name='no_sample')
    source_datasets_1 = [DataSet(conn=source_conn,
                                 exp_id=source_exp_1.exp_id) for i in range(5)]
    source_datasets_2 = [DataSet(conn=source_conn,
                                 exp_id=source_exp_2.exp_id) for i in range(5)]
    source_datasets = source_datasets_1 + source_datasets_2

    for ds in source_datasets:
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{name: 0.0 for name in some_interdeps[1].names}])
        ds.mark_completed()

    # now let's insert only some of the datasets
    # and verify that the ids and counters are set correctly
    for ds in source_datasets[-3:]:
        extract_runs_into_db(ds.conn.path_to_dbfile,
                             target_conn.path_to_dbfile, ds.run_id)

    loaded_datasets = [load_by_run_spec(captured_run_id=i, conn=target_conn)
                       for i in range(8, 11)]
    expected_run_ids = [1, 2, 3]
    expected_captured_run_ids = [8, 9, 10]
    expected_counter = [1, 2, 3]
    expected_captured_counter = [3, 4, 5]

    for ds, eri, ecri, ec, ecc in zip(loaded_datasets,
                                      expected_run_ids,
                                      expected_captured_run_ids,
                                      expected_counter,
                                      expected_captured_counter):
        assert ds.run_id == eri
        assert ds.captured_run_id == ecri
        assert ds.counter == ec
        assert ds.captured_counter == ecc

    exp = load_experiment_by_name('exp2', conn=target_conn)

    # add additional runs and verify that the ids and counters increase as
    # expected
    new_datasets = [DataSet(conn=target_conn,
                            exp_id=exp.exp_id) for i in range(3)]

    for ds in new_datasets:
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{name: 0.0 for name in some_interdeps[1].names}])
        ds.mark_completed()

    expected_run_ids = [4, 5, 6]
    expected_captured_run_ids = [11, 12, 13]
    expected_counter = [4, 5, 6]
    expected_captured_counter = [6, 7, 8]

    for ds, eri, ecri, ec, ecc in zip(new_datasets,
                                      expected_run_ids,
                                      expected_captured_run_ids,
                                      expected_counter,
                                      expected_captured_counter):
        assert ds.run_id == eri
        assert ds.captured_run_id == ecri
        assert ds.counter == ec
        assert ds.captured_counter == ecc


def test_old_versions_not_touched(two_empty_temp_db_connections,
                                  some_interdeps):

    source_conn, target_conn = two_empty_temp_db_connections

    target_path = path_to_dbfile(target_conn)
    source_path = path_to_dbfile(source_conn)

    _, new_v = get_db_version_and_newest_available_version(source_path)

    fixturepath = os.sep.join(qcodes.tests.dataset.__file__.split(os.sep)[:-1])
    fixturepath = os.path.join(fixturepath,
                               'fixtures', 'db_files', 'version2',
                               'some_runs.db')
    if not os.path.exists(fixturepath):
        pytest.skip("No db-file fixtures found. You can generate test db-files"
                    " using the scripts in the legacy_DB_generation folder")

    # First test that we cannot use an old version as source

    with raise_if_file_changed(fixturepath):
        with pytest.warns(UserWarning) as warning:
            extract_runs_into_db(fixturepath, target_path, 1)
            expected_mssg = ('Source DB version is 2, but this '
                             f'function needs it to be in version {new_v}. '
                             'Run this function again with '
                             'upgrade_source_db=True to auto-upgrade '
                             'the source DB file.')
            assert warning[0].message.args[0] == expected_mssg

    # Then test that we cannot use an old version as target

    # first create a run in the new version source
    source_exp = Experiment(conn=source_conn)
    source_ds = DataSet(conn=source_conn, exp_id=source_exp.exp_id)

    source_ds.set_interdependencies(some_interdeps[1])

    source_ds.mark_started()
    source_ds.add_results([{name: 0.0
                            for name in some_interdeps[1].names}])
    source_ds.mark_completed()

    with raise_if_file_changed(fixturepath):
        with pytest.warns(UserWarning) as warning:
            extract_runs_into_db(source_path, fixturepath, 1)
            expected_mssg = ('Target DB version is 2, but this '
                             f'function needs it to be in version {new_v}. '
                             'Run this function again with '
                             'upgrade_target_db=True to auto-upgrade '
                             'the target DB file.')
            assert warning[0].message.args[0] == expected_mssg


def test_experiments_with_NULL_sample_name(two_empty_temp_db_connections,
                                           some_interdeps):
    """
    In older API versions (corresponding to DB version 3),
    users could get away with setting the sample name to None

    This test checks that such an experiment gets correctly recognised and
    is thus not ever re-inserted into the target DB
    """
    source_conn, target_conn = two_empty_temp_db_connections
    source_exp_1 = Experiment(conn=source_conn, name='null_sample_name')

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    # make 5 runs in experiment

    exp_1_run_ids = []
    for _ in range(5):

        source_dataset = DataSet(conn=source_conn, exp_id=source_exp_1.exp_id)
        exp_1_run_ids.append(source_dataset.run_id)

        source_dataset.set_interdependencies(some_interdeps[1])
        source_dataset.mark_started()

        for val in range(10):
            source_dataset.add_results([{name: val
                                         for name in some_interdeps[1].names}])
        source_dataset.mark_completed()

    sql = """
          UPDATE experiments
          SET sample_name = NULL
          WHERE exp_id = 1
          """
    source_conn.execute(sql)
    source_conn.commit()

    assert source_exp_1.sample_name is None

    extract_runs_into_db(source_path, target_path, 1, 2, 3, 4, 5)

    assert len(get_experiments(target_conn)) == 1

    extract_runs_into_db(source_path, target_path, 1, 2, 3, 4, 5)

    assert len(get_experiments(target_conn)) == 1

    assert len(Experiment(exp_id=1, conn=target_conn)) == 5


def test_integration_station_and_measurement(two_empty_temp_db_connections,
                                             inst):
    """
    An integration test where the runs in the source DB file are produced
    with the Measurement object and there is a Station as well
    """
    source_conn, target_conn = two_empty_temp_db_connections
    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    source_exp = Experiment(conn=source_conn)

    # Set up measurement scenario
    station = Station(inst)

    meas = Measurement(exp=source_exp, station=station)
    meas.register_parameter(inst.back)
    meas.register_parameter(inst.plunger)
    meas.register_parameter(inst.cutter, setpoints=(inst.back, inst.plunger))

    with meas.run() as datasaver:
        for back_v in [1, 2, 3]:
            for plung_v in [-3, -2.5, 0]:
                datasaver.add_result((inst.back, back_v),
                                     (inst.plunger, plung_v),
                                     (inst.cutter, back_v+plung_v))

    extract_runs_into_db(source_path, target_path, 1)

    target_ds = DataSet(conn=target_conn, run_id=1)

    assert datasaver.dataset.the_same_dataset_as(target_ds)


def test_atomicity(two_empty_temp_db_connections, some_interdeps):
    """
    Test the atomicity of the transaction by extracting and inserting two
    runs where the second one is not completed. The not completed error must
    roll back any changes to the target
    """
    source_conn, target_conn = two_empty_temp_db_connections

    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    # The target file must exist for us to be able to see whether it has
    # changed
    Path(target_path).touch()

    source_exp = Experiment(conn=source_conn)
    source_ds_1 = DataSet(conn=source_conn, exp_id=source_exp.exp_id)
    source_ds_2 = DataSet(conn=source_conn, exp_id=source_exp.exp_id)

    for ds in (source_ds_1, source_ds_2):
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{name: 2.1
                         for name in some_interdeps[1].names}])

    # importantly, source_ds_2 is NOT marked as completed
    source_ds_1.mark_completed()

    # now check that the target file is untouched
    with raise_if_file_changed(target_path):
        # although the not completed error is a ValueError, we get the
        # RuntimeError from SQLite
        with pytest.raises(RuntimeError):
            extract_runs_into_db(source_path, target_path, 1, 2)


def test_column_mismatch(two_empty_temp_db_connections, some_interdeps, inst):
    """
    Test insertion of runs with no metadata and no snapshot into a DB already
    containing a run that has both
    """

    source_conn, target_conn = two_empty_temp_db_connections
    source_path = path_to_dbfile(source_conn)
    target_path = path_to_dbfile(target_conn)

    target_exp = Experiment(conn=target_conn)

    # Set up measurement scenario
    station = Station(inst)

    meas = Measurement(exp=target_exp, station=station)
    meas.register_parameter(inst.back)
    meas.register_parameter(inst.plunger)
    meas.register_parameter(inst.cutter, setpoints=(inst.back, inst.plunger))

    with meas.run() as datasaver:
        for back_v in [1, 2, 3]:
            for plung_v in [-3, -2.5, 0]:
                datasaver.add_result((inst.back, back_v),
                                     (inst.plunger, plung_v),
                                     (inst.cutter, back_v+plung_v))
    datasaver.dataset.add_metadata('meta_tag', 'meta_value')

    Experiment(conn=source_conn)
    source_ds = DataSet(conn=source_conn)
    source_ds.set_interdependencies(some_interdeps[1])

    source_ds.mark_started()
    source_ds.add_results([{name: 2.1
                            for name in some_interdeps[1].names}])
    source_ds.mark_completed()

    extract_runs_into_db(source_path, target_path, 1)

    # compare
    target_copied_ds = DataSet(conn=target_conn, run_id=2)

    assert target_copied_ds.the_same_dataset_as(source_ds)



from typing import Any

from ..param_spec import ParamSpec
from .rundescribertypes import InterDependenciesDict


class InterDependencies:
    """
    Object containing the ParamSpecs of a given run
    """

    def __init__(self, *paramspecs: ParamSpec) -> None:

        for paramspec in paramspecs:
            if not isinstance(paramspec, ParamSpec):
                raise ValueError('Got invalid input. All paramspecs must be '
                                 f'ParamSpecs, but {paramspec} is of type '
                                 f'{type(paramspec)}.')

        self.paramspecs = paramspecs

    def __repr__(self) -> str:
        output = self.__class__.__name__
        tojoin = (str(paramspec) for paramspec in self.paramspecs)
        output += f'({", ".join(tojoin)})'
        return output

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, InterDependencies):
            return False
        ours = sorted(self.paramspecs, key=lambda ps: ps.name)
        theirs = sorted(other.paramspecs, key=lambda ps: ps.name)
        if not ours == theirs:
            return False
        return True

    def _to_dict(self) -> InterDependenciesDict:
        """
        Return a dictionary representation of this object instance
        """

        return {'paramspecs': tuple(ps._to_dict() for ps in self.paramspecs)}


    @classmethod
    def _from_dict(cls, ser: InterDependenciesDict) -> 'InterDependencies':
        """
        Create an InterDependencies object from a dictionary
        """
        paramspecs = [ParamSpec._from_dict(sps) for sps in ser['paramspecs']]
        idp = cls(*paramspecs)
        return idp



import hypothesis.strategies as hst
import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings
from numpy.testing import assert_allclose

import qcodes as qc
from qcodes.dataset.data_export import _get_data_from_ds, get_data_by_id
from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.measurements import Measurement
from qcodes.utils.deprecate import QCoDeSDeprecationWarning


def test_get_data_by_id_order(dataset):
    """
    Test that the added values of setpoints end up associated with the correct
    setpoint parameter, irrespective of the ordering of those setpoint
    parameters
    """
    indepA = ParamSpecBase('indep1', "numeric")
    indepB = ParamSpecBase('indep2', "numeric")
    depAB = ParamSpecBase('depAB', "numeric")
    depBA = ParamSpecBase('depBA', "numeric")

    idps = InterDependencies_(
        dependencies={depAB: (indepA, indepB), depBA: (indepB, indepA)})

    dataset.set_interdependencies(idps)

    dataset.mark_started()

    dataset.add_results([{'depAB': 12,
                          'indep2': 2,
                          'indep1': 1}])

    dataset.add_results([{'depBA': 21,
                          'indep2': 2,
                          'indep1': 1}])
    dataset.mark_completed()

    with pytest.warns(QCoDeSDeprecationWarning):
        data1 = get_data_by_id(dataset.run_id)

    data2 = _get_data_from_ds(dataset)

    for data in (data1, data2):
        data_dict = {el['name']: el['data'] for el in data[0]}
        assert data_dict['indep1'] == 1
        assert data_dict['indep2'] == 2

        data_dict = {el['name']: el['data'] for el in data[1]}
        assert data_dict['indep1'] == 1
        assert data_dict['indep2'] == 2


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multidimarrayparameter_as_array(
        SpectrumAnalyzer,
        bg_writing
):
    array_param = SpectrumAnalyzer.multidimspectrum
    meas = Measurement()
    meas.register_parameter(array_param, paramtype='array')
    assert len(meas.parameters) == 4
    inserted_data = array_param.get()
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((array_param, inserted_data))

    expected_shape = (1, 100, 50, 20)

    datadicts = _get_data_from_ds(datasaver.dataset)
    assert len(datadicts) == 1
    for datadict_list in datadicts:
        assert len(datadict_list) == 4
        for datadict in datadict_list:

            datadict['data'].shape = (np.prod(expected_shape),)
            if datadict['name'] == "dummy_SA_Frequency0":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[0])
                expected_data = np.repeat(temp_data,
                                          expected_shape[2] * expected_shape[3])
            if datadict['name'] == "dummy_SA_Frequency1":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[1])
                expected_data = np.tile(np.repeat(temp_data, expected_shape[3]),
                                        expected_shape[1])
            if datadict['name'] == "dummy_SA_Frequency2":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[2])
                expected_data = np.tile(temp_data,
                                        expected_shape[1] * expected_shape[2])
            if datadict['name'] == "dummy_SA_multidimspectrum":
                expected_data = inserted_data.ravel()
            assert_allclose(datadict['data'], expected_data)


@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.usefixtures("experiment")
def test_datasaver_multidimarrayparameter_as_numeric(SpectrumAnalyzer,
                                                     bg_writing):
    """
    Test that storing a multidim Array parameter as numeric unravels the
    parameter as expected.
    """

    array_param = SpectrumAnalyzer.multidimspectrum
    meas = Measurement()
    meas.register_parameter(array_param, paramtype='numeric')
    expected_shape = array_param.shape
    dims = len(array_param.shape)
    assert len(meas.parameters) == dims + 1

    points_expected = np.prod(array_param.npts)
    inserted_data = array_param.get()
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((array_param, inserted_data))

    assert datasaver.points_written == points_expected

    datadicts = _get_data_from_ds(datasaver.dataset)
    assert len(datadicts) == 1
    for datadict_list in datadicts:
        assert len(datadict_list) == 4
        for datadict in datadict_list:

            datadict['data'].shape = (np.prod(expected_shape),)
            if datadict['name'] == "dummy_SA_Frequency0":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[0])
                expected_data = np.repeat(temp_data,
                                          expected_shape[1] * expected_shape[2])
            if datadict['name'] == "dummy_SA_Frequency1":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[1])
                expected_data = np.tile(np.repeat(temp_data, expected_shape[2]),
                                        expected_shape[0])
            if datadict['name'] == "dummy_SA_Frequency2":
                temp_data = np.linspace(array_param.start,
                                        array_param.stop,
                                        array_param.npts[2])
                expected_data = np.tile(temp_data,
                                        expected_shape[0] * expected_shape[1])
            if datadict['name'] == "dummy_SA_multidimspectrum":
                expected_data = inserted_data.ravel()
            assert_allclose(datadict['data'], expected_data)


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=500))
@pytest.mark.parametrize("bg_writing", [True, False])
@pytest.mark.parametrize("storage_type", ['numeric', 'array'])
@pytest.mark.usefixtures("experiment")
def test_datasaver_array_parameters_channel(channel_array_instrument,
                                            DAC, N, storage_type,
                                            bg_writing):
    array_param = channel_array_instrument.A.dummy_array_parameter
    meas = Measurement()
    meas.register_parameter(DAC.ch1)
    meas.register_parameter(array_param, setpoints=[DAC.ch1], paramtype=storage_type)

    M = array_param.shape[0]
    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in np.linspace(0, 0.01, N):
            datasaver.add_result((DAC.ch1, set_v),
                                 (array_param, array_param.get()))

    datadicts = _get_data_from_ds(datasaver.dataset)
    # one dependent parameter
    assert len(datadicts) == 1
    datadicts = datadicts[0]
    assert len(datadicts) == len(meas.parameters)
    for datadict in datadicts:
        if storage_type == "array":
            assert datadict["data"].shape == (N, M)
        else:
            assert datadict["data"].shape == (N * M,)


@settings(max_examples=5, deadline=None,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(N=hst.integers(min_value=5, max_value=500))
@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_array_parameters_array(channel_array_instrument, DAC, N,
                                          bg_writing):
    """
    Test that storing array parameters inside a loop works as expected
    """
    storage_type = "array"
    array_param = channel_array_instrument.A.dummy_array_parameter
    dependency_name = 'dummy_channel_inst_ChanA_array_setpoint_param_this_setpoint'

    # Now for a real measurement

    meas = Measurement()

    meas.register_parameter(DAC.ch1, paramtype='numeric')
    meas.register_parameter(array_param, setpoints=[DAC.ch1], paramtype=storage_type)

    assert len(meas.parameters) == 3

    M = array_param.shape[0]
    dac_datapoints = np.linspace(0, 0.01, N)
    with meas.run(write_in_background=bg_writing) as datasaver:
        for set_v in dac_datapoints:
            datasaver.add_result((DAC.ch1, set_v),
                                 (array_param, array_param.get()))

    datadicts = _get_data_from_ds(datasaver.dataset)
    # one dependent parameter
    assert len(datadicts) == 1
    datadicts = datadicts[0]
    assert len(datadicts) == len(meas.parameters)
    for datadict in datadicts:
        if datadict['name'] == 'dummy_dac_ch1':
            expected_data = np.repeat(dac_datapoints, M).reshape(N, M)
        if datadict['name'] == dependency_name:
            expected_data = np.tile(np.linspace(5, 9, 5), (N, 1))
        if datadict['name'] == 'dummy_channel_inst_ChanA_dummy_array_parameter':
            expected_data = np.empty((N, M))
            expected_data[:] = 2.
        assert_allclose(datadict['data'], expected_data)

        assert datadict["data"].shape == (N, M)


@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_multidim_array(experiment, bg_writing):
    """
    Test that inserting multidim parameters as arrays works as expected
    """
    meas = Measurement(experiment)
    size1 = 10
    size2 = 15

    data_mapping = {name: i for i, name in
                    zip(range(4), ['x1', 'x2', 'y1', 'y2'])}

    x1 = qc.ManualParameter('x1')
    x2 = qc.ManualParameter('x2')
    y1 = qc.ManualParameter('y1')
    y2 = qc.ManualParameter('y2')

    meas.register_parameter(x1, paramtype='array')
    meas.register_parameter(x2, paramtype='array')
    meas.register_parameter(y1, setpoints=[x1, x2], paramtype='array')
    meas.register_parameter(y2, setpoints=[x1, x2], paramtype='array')
    data = np.random.rand(4, size1, size2)
    expected = {'x1': data[0, :, :],
                'x2': data[1, :, :],
                'y1': data[2, :, :],
                'y2': data[3, :, :]}
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((str(x1), expected['x1']),
                             (str(x2), expected['x2']),
                             (str(y1), expected['y1']),
                             (str(y2), expected['y2']))

    datadicts = _get_data_from_ds(datasaver.dataset)
    assert len(datadicts) == 2
    for datadict_list in datadicts:
        assert len(datadict_list) == 3
        for datadict in datadict_list:
            dataindex = data_mapping[datadict["name"]]
            expected_data = data[dataindex : dataindex + 1, :, :]
            assert_allclose(datadict["data"], expected_data)

            assert datadict["data"].shape == (1, size1, size2)


@pytest.mark.parametrize("bg_writing", [True, False])
def test_datasaver_multidim_numeric(experiment, bg_writing):
    """
    Test that inserting multidim parameters as numeric works as expected
    """
    meas = Measurement(experiment)
    size1 = 10
    size2 = 15
    x1 = qc.ManualParameter('x1')
    x2 = qc.ManualParameter('x2')
    y1 = qc.ManualParameter('y1')
    y2 = qc.ManualParameter('y2')

    data_mapping = {name: i for i, name in
                    zip(range(4), ['x1', 'x2', 'y1', 'y2'])}

    meas.register_parameter(x1, paramtype='numeric')
    meas.register_parameter(x2, paramtype='numeric')
    meas.register_parameter(y1, setpoints=[x1, x2], paramtype='numeric')
    meas.register_parameter(y2, setpoints=[x1, x2], paramtype='numeric')
    data = np.random.rand(4, size1, size2)
    with meas.run(write_in_background=bg_writing) as datasaver:
        datasaver.add_result((str(x1), data[0, :, :]),
                             (str(x2), data[1, :, :]),
                             (str(y1), data[2, :, :]),
                             (str(y2), data[3, :, :]))

    datadicts = _get_data_from_ds(datasaver.dataset)
    assert len(datadicts) == 2
    for datadict_list in datadicts:
        assert len(datadict_list) == 3
        for datadict in datadict_list:
            dataindex = data_mapping[datadict['name']]
            expected_data = data[dataindex, :, :].ravel()
            assert_allclose(datadict['data'], expected_data)

            assert datadict['data'].shape == (size1 * size2,)



import pytest
from pyvisa.errors import VisaIOError
from qcodes.instrument_drivers.Keysight.keysight_b220x import KeysightB220X


@pytest.fixture
def uut():
    try:
        resource_name = 'insert_Keysight_B2200_VISA_resource_name_here'
        instance = KeysightB220X('switch_matrix',
                                 address=resource_name)
    except (ValueError, VisaIOError):
        # Either there is no VISA lib installed or there was no real
        # instrument found at the specified address => use simulated instrument
        import qcodes.instrument.sims as sims
        path_to_yaml = sims.__file__.replace('__init__.py',
                                             'keysight_b220x.yaml')

        instance = KeysightB220X('switch_matrix',
                                 address='GPIB::1::INSTR',
                                 visalib=path_to_yaml + '@sim'
                                 )

    instance.get_status()
    instance.clear_status()
    instance.reset()

    yield instance

    instance.close()


def test_idn_command(uut):
    assert "AGILENT" in uut.IDN()['vendor']
    assert 0 == uut.get_status()


def test_connect(uut):
    uut.connect(2, 48)
    assert 0 == uut.get_status()


def test_connect_throws_at_invalid_channel_number(uut):
    with pytest.raises(ValueError):
        uut.connect(2, 49)
    with pytest.raises(ValueError):
        uut.connect(2, 0)
    with pytest.raises(ValueError):
        uut.connect(0, 10)
    with pytest.raises(ValueError):
        uut.connect(15, 10)


def test_connect_emits_warning_on_statusbyte_not_null(uut):
    # some tricks are used to trigger an instrument error both in the
    # simulation as well as in the real instrument:
    # 1. with gnd mode enabled, it is illegal to connect to input channel 12,
    #  which will raise an instrument error in the physical instrument,
    # but not in the simulated instrument.
    # 2. The simulated instrument only accepts pre-defined channels in the
    # connect command. here a channel is used that was not pre-defined which
    # will cause an instrument error in the simulated instrument, but not in
    # the physical instrument

    uut.gnd_mode(True)
    with pytest.warns(UserWarning):
        uut.connect(12, 33)

        # The simulated instrument does not reset the settings to default
        # values, so gnd mode is explicitly disabled here:
        uut.gnd_mode(False)


def test_disconnect_throws_at_invalid_channel_number(uut):
    with pytest.raises(ValueError):
        uut.disconnect(2, 49)
    with pytest.raises(ValueError):
        uut.disconnect(2, 0)
    with pytest.raises(ValueError):
        uut.disconnect(0, 10)
    with pytest.raises(ValueError):
        uut.disconnect(15, 10)


def test_connections(uut):
    uut.connect(2, 48)
    uut.connect(10, 12)
    assert {(2, 48), (10, 12)} == uut.connections()


def test_to_channel_list(uut):
    assert '(@00345,01109)' == uut.to_channel_list([(3, 45), (11, 9)])


def test_connect_paths(uut):
    uut.disconnect_all()
    uut.connect_paths([(3, 45), (11, 9)])
    assert 0 == uut.get_status()


def test_disconnect_paths(uut):
    uut.connect_paths([(3, 45), (11, 9)])
    uut.disconnect_paths([(3, 45), (11, 9)])
    assert 0 == uut.get_status()


def test_disconnect_all(uut):
    uut.connect(2, 48)
    uut.disconnect_all()
    assert 0 == uut.get_status()


def test_disconnect(uut):
    uut.connect(2, 48)
    assert 0 == uut.get_status()
    uut.disconnect(2, 48)
    assert 0 == uut.get_status()
    uut.disconnect(3, 22)
    assert 0 == uut.get_status()


@pytest.mark.filterwarnings("ignore:When going")
def test_connection_rule(uut):
    uut.connection_rule('single')
    assert 0 == uut.get_status()
    assert 'single' == uut.connection_rule()
    assert 0 == uut.get_status()


def test_connection_rule_emits_warning_when_going_from_free_to_single(uut):
    uut.connection_rule(
        'free')  # uut should already be in free mode after reset

    with pytest.warns(UserWarning):
        uut.connection_rule('single')


def test_connection_sequence(uut):
    assert 'bbm' == uut.connection_sequence()
    assert 0 == uut.get_status()
    uut.connection_sequence('mbb')
    assert 0 == uut.get_status()
    assert 'mbb' == uut.connection_sequence()


def test_bias_disable_all_outputs(uut):
    uut.bias_disable_all_outputs()
    assert 0 == uut.get_status()


def test_bias_disable_ouput(uut):
    uut.bias_disable_output(1)
    assert 0 == uut.get_status()


def test_bias_enable_all_outputs(uut):
    uut.bias_enable_all_outputs()
    assert 0 == uut.get_status()


def test_bias_enable_output(uut):
    uut.bias_enable_output(1)
    assert 0 == uut.get_status()


def test_bias_input_port(uut):
    assert 10 == uut.bias_input_port()
    uut.bias_input_port(9)
    assert 9 == uut.bias_input_port()
    assert 0 == uut.get_status()


def test_bias_mode(uut):
    uut.bias_mode(True)
    assert uut.bias_mode()
    assert 0 == uut.get_status()


def test_gnd_disable_all_outputs(uut):
    uut.gnd_disable_all_outputs()
    assert 0 == uut.get_status()


def test_gnd_disable_output(uut):
    uut.gnd_disable_output(1)
    assert 0 == uut.get_status()


def test_gnd_enable_all_outputs(uut):
    uut.gnd_enable_all_outputs()
    assert 0 == uut.get_status()


def test_gnd_enable_output(uut):
    uut.gnd_enable_output(1)
    assert 0 == uut.get_status()


def test_gnd_input_port(uut):
    assert 12 == uut.gnd_input_port()
    uut.gnd_input_port(5)
    assert 5 == uut.gnd_input_port()
    assert 0 == uut.get_status()


def test_gnd_mode(uut):
    assert not uut.gnd_mode()
    uut.gnd_mode(True)
    assert uut.gnd_mode()
    assert 0 == uut.get_status()


def test_unused_inputs(uut):
    uut.unused_inputs()
    assert 0 == uut.get_status()

    uut.unused_inputs([3])
    assert 0 == uut.get_status()
    assert [3] == uut.unused_inputs()

    uut.unused_inputs([5, 6, 7, 8])
    assert 0 == uut.get_status()
    assert [5, 6, 7, 8] == uut.unused_inputs()


def test_couple_mode(uut):
    assert not uut.couple_mode()
    uut.couple_mode(True)
    assert uut.couple_mode()
    assert 0 == uut.get_status()


def test_couple_ports(uut):
    assert not uut.couple_ports()
    assert 0 == uut.get_status()

    uut.couple_ports([1])
    assert 0 == uut.get_status()
    assert [1] == uut.couple_ports()

    uut.couple_ports([1, 3, 5])
    assert 0 == uut.get_status()
    assert [1, 3, 5] == uut.couple_ports()

    with pytest.raises(ValueError):
        uut.couple_ports([2])
    with pytest.raises(ValueError):
        uut.couple_ports([2, 3])


def test_couple_port_autodetect(uut):
    uut.couple_port_autodetect()
    assert 0 == uut.get_status()


def test_get_error(uut):
    uut.get_error()
    assert 0 == uut.get_status()


class TestParseChannelList:
    @staticmethod
    def test_parse_channel_list():
        channel_list = '(@10101,10202)'
        assert {(1, 1), (2, 2)} == KeysightB220X.parse_channel_list(
            channel_list)

    @staticmethod
    def test_all_combinations_zero_padded():
        import itertools
        cards = range(5)
        inputs = range(1, 15)
        outputs = range(1, 49)

        for card, in_port, out_port in itertools.product(cards, inputs,
                                                         outputs):
            padded = f'{card:01d}{in_port:02d}{out_port:02d}'

            assert {(in_port, out_port)} == KeysightB220X.parse_channel_list(
                padded)

    @staticmethod
    def test_all_combinations_unpadded():
        import itertools
        cards = range(5)
        inputs = range(1, 15)
        outputs = range(1, 49)

        for card, in_port, out_port in itertools.product(cards, inputs,
                                                         outputs):
            padded = f'{card:01d}{in_port:02d}{out_port:02d}'
            unpadded = str(int(padded))

            assert {(in_port, out_port)} == KeysightB220X.parse_channel_list(
                unpadded)



import pytest
from qcodes.utils.validators import Strings

from .conftest import AClass, a_func

long_string = '+'.join(str(i) for i in range(100000))
danish = '\u00d8rsted F\u00e6lled'
chinese = '\u590f\u65e5\u7545\u9500\u699c\u5927\u724c\u7f8e'

strings = ['', '0', '10', '1.0e+10', 'a', 'Ja', 'Artichokes!',
           danish, chinese, long_string]

not_strings = [0, 1, 1.0e+10, bytes('', 'utf8'),
               bytes(danish, 'utf8'), bytes(chinese, 'utf8'),
               [], [1, 2, 3], {}, {'a': 1, 'b': 2},
               True, False, None, AClass, AClass(), a_func]


def test_unlimited():
    s = Strings()

    for v in strings:
        s.validate(v)

    for v in not_strings:
        with pytest.raises(TypeError):
            s.validate(v)

    assert repr(s) == '<Strings>'


def test_min():
    for min_len in [0, 1, 5, 10, 100]:
        s = Strings(min_length=min_len)
        for v in strings:
            if len(v) >= min_len:
                s.validate(v)
            else:
                with pytest.raises(ValueError):
                    s.validate(v)

        for v in not_strings:
            with pytest.raises(TypeError):
                s.validate(v)

    assert repr(s) == '<Strings len>=100>'


def test_max():
    for max_len in [1, 5, 10, 100]:
        s = Strings(max_length=max_len)
        for v in strings:
            if len(v) <= max_len:
                s.validate(v)
            else:
                with pytest.raises(ValueError):
                    s.validate(v)

    for v in not_strings:
        with pytest.raises(TypeError):
            s.validate(v)

    assert repr(s) == '<Strings len<=100>'


def test_range():
    s = Strings(1, 10)

    for v in strings:
        if 1 <= len(v) <= 10:
            s.validate(v)
        else:
            with pytest.raises(ValueError):
                s.validate(v)

    for v in not_strings:
        with pytest.raises(TypeError):
            s.validate(v)

    assert repr(s) == '<Strings 1<=len<=10>'

    # single-valued range
    assert repr(Strings(10, 10)) == '<Strings len=10>'


def test_failed_strings():
    with pytest.raises(TypeError):
        Strings(1, 2, 3)

    with pytest.raises(TypeError):
        Strings(10, 9)

    with pytest.raises(TypeError):
        Strings(max_length=0)

    with pytest.raises(TypeError):
        Strings(min_length=1e12)

    for length in [-1, 3.5, '2', None]:
        with pytest.raises(TypeError):
            Strings(max_length=length)

        with pytest.raises(TypeError):
            Strings(min_length=length)


def test_valid_values():
    val = Strings()
    for vval in val.valid_values:
        val.validate(vval)



import logging
import re
import warnings
from typing import Dict, Optional, Type, Hashable, Any

from qcodes.instrument.base import Instrument
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.utils.validators import Ints

log = logging.getLogger(__name__)


class SwitchChannelBase(InstrumentChannel):
    def __init__(self, parent: Instrument, name: str, channel_letter: str):
        """
        Args:
            parent: The instrument the channel is a part of
            name: the name of the channel
            channel_letter: channel letter ['a', 'b', 'c' or 'd'])
        """

        super().__init__(parent, name)
        self.channel_letter = channel_letter.upper()
        _chanlist = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
        self.channel_number = _chanlist.index(channel_letter)

        self.add_parameter(
            'switch',
            label=f'switch {self.channel_letter}',
            set_cmd=self._set_switch,
            get_cmd=self._get_switch,
            vals=Ints(1, 2))

    def __call__(self, *args: int) -> Optional[int]:
        if len(args) == 1:
            self.switch(args[0])
            return None
        elif len(args) == 0:
            return self.switch()
        else:
            raise RuntimeError(
                'Call channel with either one or zero arguments')

    def _set_switch(self, switch: int) -> None:
        raise NotImplementedError()

    def _get_switch(self) -> int:
        raise NotImplementedError()


class SPDT_Base(Instrument):

    CHANNEL_CLASS: Type[SwitchChannelBase]

    def add_channels(self) -> None:
        channels = ChannelList(
            self, "Channels", self.CHANNEL_CLASS, snapshotable=False)

        _chanlist = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
        self._deprecated_attributes: Dict[str, str] = {
            f'channel_{k}': k
            for k in _chanlist
        }

        _max_channel_number = self.get_number_of_channels()
        _chanlist = _chanlist[0:_max_channel_number]

        for c in _chanlist:
            channel = self.CHANNEL_CLASS(self, f'channel_{c}', c)
            channels.append(channel)
            attribute_name = f'channel_{c}'
            self.add_submodule(attribute_name, channel)
            self.add_submodule(c, channel)
            self._deprecated_attributes[attribute_name] = c
        channels.lock()
        self.add_submodule('channels', channels)

    def all(self, switch_to: int) -> None:
        for c in self.channels:
            c.switch(switch_to)

    def __getattr__(self, key: str) -> Any:
        if key in self._deprecated_attributes:
            warnings.warn(
                ("Using '{}' is deprecated and will be removed in future" +
                 "releases. Use '{}' instead").
                format(key, self._deprecated_attributes[key]), UserWarning)
        return super().__getattr__(key)

    def get_number_of_channels(self) -> int:
        model = self.get_idn()['model']
        if model is None:
            raise RuntimeError(
                'The driver could not get model information for the device, '
                'it might not be supported.'
            )
        model_parts = model.split('-')
        if len(model_parts) < 2:
            raise RuntimeError(
                ('The driver could not determine the number of channels of ' +
                 'the model \'{}\', it might not be supported').format(model)
            )
        if model_parts[0] not in ('RC', 'USB'):
            log.warning(
                ('The model with the name \'{}\' might not be supported by' +
                 ' the driver').format(model))
        match = re.match('^[0-9]*', model_parts[1])
        if match is None:
            raise RuntimeError(
                'The driver could not determine the number of channels of' +
                f' the model \'{model}\', it might not be supported'
            )
        channels = match[0]
        if not channels:
            raise RuntimeError(
                'The driver could not determine the number of channels of' +
                f' the model \'{model}\', it might not be supported'
            )
        return int(channels)



import re
import time
from datetime import datetime


class MockAMI430:
    states = {
        "RAMPING to target field/current": "1",
        "HOLDING at the target field/current": "2",
        "PAUSED": "3",
        "Ramping in MANUAL UP mode": "4",
        "Ramping in MANUAL DOWN mode": "5",
        "ZEROING CURRENT (in progress)": "6",
        "Quench detected": "7",
        "At ZERO current": "8",
        "Heating persistent switch": "9",
        "Cooling persistent switch": "10"
    }

    field_units = {
        "tesla": "1",
        "kilogauss": "0"
    }

    ramp_rate_units = {
        "A/s": "0",
        "A/min": "1"
    }

    quench_state = {False: "0", True: "1"}

    def __init__(self, name):

        self.name = name
        self.log_messages = []

        self._field_mag = 0
        self._field_target = 0
        self._state = MockAMI430.states["HOLDING at the target field/current"]

        self.handlers = {
            "RAMP:RATE:UNITS": {
                "get": MockAMI430.ramp_rate_units["A/s"],
                "set": None
            },
            "FIELD:UNITS": {
                "get": MockAMI430.field_units["tesla"],
                "set": None
            },
            "*IDN": {
                "get": "v0.1 Mock",
                "set": None
            },
            "STATE": {
                "get": self._getter("_state"),
                "set": self._setter("_state")
            },
            "FIELD:MAG": {
                "get": self._getter("_field_mag"),
                "set": None
            },
            "QU": {
                "get": MockAMI430.quench_state[False],  # We are never in a quenching state so always return the
                # same value
                "set": None
            },
            "PERS": {
                "get": "0",
                "set": None
            },
            "PAUSE": {
                "get": self._is_paused,
                "set": self._do_pause
            },
            "CONF:FIELD:TARG": {
                "get": None,  # To get the field target, send a message "FIELD:TARG?"
                "set": self._setter("_field_target")
            },
            "FIELD:TARG": {
                "get": self._getter("_field_target"),
                "set": None
            },
            "PS": {
                "get": "0",  # The heater is off
                "set": None
            },
            "RAMP": {
                "set": self._do_ramp,
                "get": None
            },
            "RAMP:RATE:CURRENT": {
                "get": "0.1000,50.0000",
                "set": None
            },
            "COIL": {
                "get": "1",
                "set": None
            }
        }

    @staticmethod
    def message_parser(gs, msg_str, key):
        """
        * If gs = "get":
        Let suppose key = "RAMP:RATE:UNITS", then if we get msg_str =
        "RAMP:RATE:UNITS?" then match will be True and args = None. If
        msg_str = "RAMP:RATE:UNITS:10?" then match = True and args =
        "10". On the other hand if key = "RAMP" then both
        "RAMP:RATE:UNITS?" and "RAMP:RATE:UNITS:10?" will cause match
        to be False

        * If gs = "set"
        If key = "STATE" and msg_str = "STATE 2,1" then match = True
        and args = "2,1". If key="STATE" and msg_str = STATE:ELSE 2,1
        then match is False.

        Consult [1] for a complete description of the AMI430 protocol.

        [1]
        http://www.americanmagnetics.com/support/manuals/mn-4Q06125PS-430.pdf

        Args: gs (string): "get", or "set" msg_str (string): the
            message string the mock instrument gets.  key (string):
            one of the keys in self.handlers

        Returns: match (bool): if the key and the msg_str match, then
            match = True args (string): if any arguments are present
            in the message string these will be passed along. This is
            always None when match = False

        """
        # If the message string matches a key exactly we have a match
        # with no arguments
        if msg_str == key:
            return True, None

        # We use regular expressions to find out if the message string
        # and the key match. We need to replace reserved regular
        # expression characters in the key. For instance replace
        # "*IDN" with "\*IDN".
        reserved_re_characters = r"\^${}[]().*+?|<>-&"
        for c in reserved_re_characters:
            key = key.replace(c, fr"\{c}")

        # Get and set messages use different regular expression
        s = {"get": r"(:[^:]*)?\?$", "set": "([^:]+)"}[gs]
        # patterns to determine a match
        search_string = "^" + key + s
        r = re.search(search_string, msg_str)
        match = r is not None

        args = None
        if match:
            args = r.groups()[0]
            if args is not None:
                args = args.strip(":")

        return match, args

    def _getter(self, attribute):
        return lambda _: getattr(self, attribute)

    def _setter(self, attribute):
        return lambda value: setattr(self, attribute, value)

    def _log(self, msg):

        now = datetime.now()
        log_line = "[{}] {}: {}".format(now.strftime("%d:%m:%Y-%H:%M:%S.%f"),
                                        self.name, msg)
        self.log_messages.append(log_line)

    def _handle_messages(self, msg):
        """
        Args:
            msg (string): a message received through the socket
                communication layer

        Returns:
            rval (string or None): If the type of message requests a
                value (a get message) then this value is returned by this
                function. A set message will return a None value.
        """

        # A "get" message ends with a "?" and will invoke the get
        # part of the handler defined in self.handlers.
        gs = {True: "get", False: "set"}[msg.endswith("?")]

        rval = None
        handler = None

        # Find which handler is suitable to handle the message
        for key in self.handlers:
            match, args = MockAMI430.message_parser(gs, msg, key)
            if not match:
                continue

            handler = self.handlers[key][gs]
            if callable(handler):
                rval = handler(args)
            else:
                rval = handler

            break

        if handler is None:
            self._log(f"Command {msg} unknown")

        return rval

    def _do_pause(self, _):
        self._state = MockAMI430.states["PAUSED"]

    def _is_paused(self):
        return self._state == MockAMI430.states["PAUSED"]

    def _do_ramp(self, _):
        self._log(f"Ramping to {self._field_target}")
        self._state = MockAMI430.states["RAMPING to target field/current"]
        time.sleep(0.1)  # Lets pretend to be ramping for a bit
        self._field_mag = self._field_target
        self._state = MockAMI430.states["HOLDING at the target field/current"]

    def get_log_messages(self):
        return self.log_messages

    def ask(self, msg):
        return self._handle_messages(msg)

    def write(self, msg):
        self._handle_messages(msg)



import re

import hypothesis.strategies as hst
from hypothesis import HealthCheck, given, settings
import hypothesis.extra.numpy as hypnumpy
import pytest
import numpy as np

import qcodes as qc
from qcodes.dataset.measurements import DataSaver, Measurement
from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.data_export import load_by_id


def test_string_via_dataset(experiment):
    """
    Test that we can save text into database via DataSet API
    """
    p = ParamSpecBase("p", "text")

    test_set = qc.new_data_set("test-dataset")
    idps = InterDependencies_(standalones=(p,))
    test_set.set_interdependencies(idps)
    test_set.mark_started()

    test_set.add_results([{"p": "some text"}])

    test_set.mark_completed()

    assert test_set.get_parameter_data()["p"]["p"] == [["some text"]]


def test_string_via_datasaver(experiment):
    """
    Test that we can save text into database via DataSaver API
    """
    p = ParamSpecBase(name="p", paramtype="text")

    test_set = qc.new_data_set("test-dataset")
    idps = InterDependencies_(standalones=(p,))
    test_set.set_interdependencies(idps)
    test_set.mark_started()

    idps = InterDependencies_(standalones=(p,))

    data_saver = DataSaver(
        dataset=test_set, write_period=0, interdeps=idps)

    data_saver.add_result(("p", "some text"))
    data_saver.flush_data_to_database()

    assert test_set.get_parameter_data()["p"]["p"] == np.array(["some text"])


def test_string(experiment):
    """
    Test that we can save text into database via Measurement API
    """
    p = qc.Parameter('p', label='String parameter', unit='', get_cmd=None,
                     set_cmd=None, initial_value='some text')

    meas = Measurement(experiment)
    meas.register_parameter(p, paramtype='text')

    with meas.run() as datasaver:
        datasaver.add_result((p, "some text"))

    test_set = load_by_id(datasaver.run_id)

    assert test_set.get_parameter_data()["p"]["p"] == np.array(["some text"])


def test_string_with_wrong_paramtype(experiment):
    """
    Test that an exception occurs when saving string data if when registering a
    string parameter the paramtype was not set to 'text'
    """
    p = qc.Parameter('p', label='String parameter', unit='', get_cmd=None,
                     set_cmd=None, initial_value='some text')

    meas = Measurement(experiment)
    # intentionally forget `paramtype='text'`, so that the default 'numeric'
    # is used, and an exception is raised later
    meas.register_parameter(p)

    with meas.run() as datasaver:
        msg = re.escape('Parameter p is of type "numeric", but got a '
                        "result of type <U9 (some text).")
        with pytest.raises(ValueError, match=msg):
            datasaver.add_result((p, "some text"))


def test_string_with_wrong_paramtype_via_datasaver(experiment):
    """
    Test that it is not possible to add a string value for a non-text
    parameter via DataSaver object
    """
    p = ParamSpecBase("p", "numeric")

    test_set = qc.new_data_set("test-dataset")
    idps = InterDependencies_(standalones=(p,))
    test_set.set_interdependencies(idps)
    test_set.mark_started()

    idps = InterDependencies_(standalones=(p,))

    data_saver = DataSaver(
        dataset=test_set, write_period=0, interdeps=idps)

    try:
        msg = re.escape('Parameter p is of type "numeric", but got a '
                        "result of type <U9 (some text).")
        with pytest.raises(ValueError, match=msg):
            data_saver.add_result(("p", "some text"))
    finally:
        data_saver.dataset.conn.close()


def test_string_saved_and_loaded_as_numeric_via_dataset(experiment):
    """
    Test that it is possible to save a string value of a non-'text' parameter
    via DataSet API, and, importantly, to retrieve it thanks to the
    flexibility of `_convert_numeric` converter function.
    """
    p = ParamSpecBase("p", "numeric")

    test_set = qc.new_data_set("test-dataset")
    idps = InterDependencies_(standalones=(p,))
    test_set.set_interdependencies(idps)
    test_set.mark_started()

    test_set.add_results([{"p": 'some text'}])

    test_set.mark_completed()

    try:
        assert np.array(['some text']) == test_set.get_parameter_data()["p"]["p"]
    finally:
        test_set.conn.close()


def test_list_of_strings(experiment):
    """
    Test saving list of strings via DataSaver
    """
    p_values = ["X_Y", "X_X", "X_I", "I_I"]
    list_of_strings = list(np.random.choice(p_values, (10,)))

    p = qc.Parameter('p', label='String parameter', unit='', get_cmd=None,
                     set_cmd=None, initial_value='X_Y')

    meas = Measurement(experiment)
    meas.register_parameter(p, paramtype='text')

    with meas.run() as datasaver:
        datasaver.add_result((p, list_of_strings))

    test_set = load_by_id(datasaver.run_id)
    expec_data = np.array([item for item in list_of_strings])
    actual_data = test_set.get_parameter_data()["p"]["p"]

    try:
        np.testing.assert_array_equal(actual_data, expec_data)
    finally:
        test_set.conn.close()


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,),
          deadline=None)
@given(
    p_values=hypnumpy.arrays(
        dtype=hst.sampled_from(
            (hypnumpy.unicode_string_dtypes(),
             hypnumpy.byte_string_dtypes(),
             hypnumpy.timedelta64_dtypes(),
             hypnumpy.datetime64_dtypes())
        ),
        shape=hypnumpy.array_shapes()
    )
)
def test_string_and_date_data_in_array(experiment, p_values):
    p = qc.Parameter('p', label='String parameter', unit='', get_cmd=None,
                     set_cmd=None, initial_value=p_values)

    meas = Measurement(experiment)
    meas.register_parameter(p, paramtype='array')

    with meas.run() as datasaver:
        datasaver.add_result((p, p.get()))
    actual_data = datasaver.dataset.get_parameter_data()["p"]["p"]
    np.testing.assert_array_equal(
        actual_data,
        p_values.reshape((1, *p_values.shape))
    )



import enum
import logging

from os.path import normpath, expanduser, exists
from typing import Union, Optional

from qcodes import config

_log = logging.getLogger(__name__)


DATASET_CONFIG_SECTION = "dataset"
EXPORT_AUTOMATIC = "export_automatic"
EXPORT_TYPE = "export_type"
EXPORT_PATH = "export_path"
EXPORT_PREFIX = "export_prefix"


class DataExportType(enum.Enum):
    """File extensions for supported data types to export data"""
    NETCDF = "nc"
    CSV = "csv"


def set_data_export_type(export_type: str) -> None:
    """Set data export type

    Args:
        export_type: Export type to use.
            Currently supported values: netcdf, csv.
    """
    # disable file export
    if export_type is None:
        config[DATASET_CONFIG_SECTION][EXPORT_TYPE] = None

    elif hasattr(DataExportType, export_type.upper()):
        config[DATASET_CONFIG_SECTION][EXPORT_TYPE] = export_type.upper()

    else:
        _log.warning(
            "Could not set export type to '%s' because it is not supported." \
            % export_type)


def set_data_export_path(export_path: str) -> None:
    """Set path to export data to at the end of a measurement

    Args:
        export_path: An existing file path on disk

    Raises:
        ValueError: If the path does not exist, this raises an error
    """
    if not exists(export_path):
        raise ValueError(f"Cannot set export path to '{export_path}' \
        because it does not exist.")
    config[DATASET_CONFIG_SECTION][EXPORT_AUTOMATIC] = export_path


def get_data_export_type(
    export_type: Optional[Union[DataExportType, str]] = None) -> Optional[DataExportType]:
    """Get the file type for exporting data to disk at the end of
    a measurement from config

    Args:
        export_type: Export type string format to convert to DataExportType.

    Returns:
        Data export type
    """
    # If export_type is None, get value from config
    export_type = export_type or config[DATASET_CONFIG_SECTION][EXPORT_TYPE]

    if isinstance(export_type, DataExportType):
        return export_type
    elif export_type:
        if hasattr(DataExportType, export_type.upper()):
            return getattr(DataExportType, export_type.upper())
    return None


def get_data_export_automatic() -> bool:
    """Should the data be exported automatically?"""
    export_automatic = config[DATASET_CONFIG_SECTION][EXPORT_AUTOMATIC]
    return export_automatic


def get_data_export_path() -> str:
    """Get the path to export data to at the end of a measurement from config

    Returns:
        Path
    """
    return normpath(expanduser(config[DATASET_CONFIG_SECTION][EXPORT_PATH]))


def set_data_export_prefix(export_prefix: str) -> None:
    """Set the data export file name prefix to export data to at the end of
    a measurement

    Args:
        export_prefix: Prefix, e.g. "qcodes_"
    """
    config[DATASET_CONFIG_SECTION][EXPORT_PREFIX] = export_prefix


def get_data_export_prefix() -> str:
    """Get the data export file name prefix to export data to at the end of
    a measurement from config

    Returns:
        Prefix, e.g. "qcodes_"
    """
    return config[DATASET_CONFIG_SECTION][EXPORT_PREFIX]



"""DataSet class and factory functions."""

import logging
import time
from collections import OrderedDict
from copy import deepcopy
from traceback import format_exc
from typing import Any, Callable, Dict, List, Optional

import numpy as np
import xarray as xr

from qcodes.data.data_array import (
    DataArray,
    data_array_to_xarray_dictionary,
    xarray_data_array_dictionary_to_data_array,
)
from qcodes.utils.helpers import DelegateAttributes, deep_update, full_class

from .gnuplot_format import GNUPlotFormat
from .io import DiskIO
from .location import FormatLocation

log = logging.getLogger(__name__)


def new_data(location=None, loc_record=None, name=None, overwrite=False,
             io=None, **kwargs):
    """
    Create a new DataSet.

    Args:
        location (Optional[Union[str,Callable, Bool]]): If you provide a string,
            it must be an unused location in the io manager. Can also be:

            - a Callable ``location provider`` with one required parameter
              (the io manager), and one optional (``record`` dict),
              which returns a location string when called
            - ``False`` - denotes an only-in-memory temporary DataSet.

            Note that the full path to or physical location of the data is a
            combination of io + location. the default ``DiskIO`` sets the base
            directory, which this location is a relative path inside.
            Default ``DataSet.location_provider`` which is initially
            ``FormatLocation()``

        loc_record (Optional[dict]): If location is a callable, this will be
            passed to it as ``record``

        name (Optional[str]): overrides the ``name`` key in the ``loc_record``.

        overwrite (bool): Are we allowed to overwrite an existing location?
            Default False.

        io (Optional[io_manager]): base physical location of the ``DataSet``.
            Default ``DataSet.default_io`` is initially ``DiskIO('.')`` which
            says the root data directory is the current working directory, ie
            where you started the python session.

        arrays (Optional[List[qcodes.data.data_array.DataArray]): arrays to add
            to the DataSet. Can be added later with ``self.add_array(array)``.

        formatter (Optional[Formatter]): sets the file format/structure to
            write (and read) with. Default ``DataSet.default_formatter`` which
            is initially ``GNUPlotFormat()``.

        write_period (Optional[float]): seconds
            between saves to disk.
    Returns:
        A new ``DataSet`` object ready for storing new data in.
    """
    if io is None:
        io = DataSet.default_io

    if name is not None:
        if not loc_record:
            loc_record = {}
        loc_record['name'] = name

    if location is None:
        location = DataSet.location_provider

    if callable(location):
        location = location(io, record=loc_record)

    if location and (not overwrite) and io.list(location):
        raise FileExistsError('"' + location + '" already has data')

    return DataSet(location=location, io=io, **kwargs)


def load_data(location=None, formatter=None, io=None):
    """
    Load an existing DataSet.

    Args:
        location (Optional[str]): the location to load from. Default is the
            current live DataSet.
            Note that the full path to or physical location of the data is a
            combination of io + location. the default ``DiskIO`` sets the base
            directory, which this location is a relative path inside.

        formatter (Optional[Formatter]): sets the file format/structure to
            read with. Default ``DataSet.default_formatter`` which
            is initially ``GNUPlotFormat()``.

        io (Optional[io_manager]): base physical location of the ``DataSet``.
            Default ``DataSet.default_io`` is initially ``DiskIO('.')`` which
            says the root data directory is the current working directory, ie
            where you started the python session.

    Returns:
        A new ``DataSet`` object loaded with pre-existing data.
    """
    if location is False:
        raise ValueError('location=False means a temporary DataSet, '
                         'which is incompatible with load_data')

    data = DataSet(location=location, formatter=formatter, io=io)
    data.read_metadata()
    data.read()
    return data


class DataSet(DelegateAttributes):

    """
    A container for one complete measurement loop.

    May contain many individual arrays with potentially different
    sizes and dimensionalities.

    Normally a DataSet should not be instantiated directly, but through
    ``new_data`` or ``load_data``.

    Args:
        location (Union[str,bool]): A location in the io manager, or ``False``
            for an only-in-memory temporary DataSet.
            Note that the full path to or physical location of the data is a
            combination of io + location. the default ``DiskIO`` sets the base
            directory, which this location is a relative path inside.

        io (Optional[io_manager]): base physical location of the ``DataSet``.
            Default ``DataSet.default_io`` is initially ``DiskIO('.')`` which
            says the root data directory is the current working directory, ie
            where you started the python session.

        arrays (Optional[List[qcodes.data.data_array.DataArray]): arrays to add
            to the DataSet. Can be added later with ``self.add_array(array)``.

        formatter (Optional[Formatter]): sets the file format/structure to
            write (and read) with. Default ``DataSet.default_formatter`` which
            is initially ``GNUPlotFormat()``.

        write_period (Optional[float]): Only if ``mode=LOCAL``, seconds
            between saves to disk. If not ``LOCAL``, the ``DataServer`` handles
            this and generally writes more often. Use None to disable writing
            from calls to ``self.store``. Default 5.
    """

    # ie data_set.arrays['vsd'] === data_set.vsd
    delegate_attr_dicts = ['arrays']

    default_io = DiskIO('.')
    default_formatter = GNUPlotFormat()
    location_provider = FormatLocation()

    background_functions: Dict[str, Callable[..., Any]] = OrderedDict()
    """
    The value ``fn`` is a callable accepting no
    arguments, and ``key`` is a name to identify the function and help
    you attach and remove it.

    In ``DataSet.complete`` we call each of these periodically, in the
    order that they were attached.

    Note that because this is a class attribute, the functions will
    apply to every DataSet. If you want specific functions for one
    DataSet you can override this with an instance attribute.
    """

    def __init__(self, location=None, arrays=None, formatter=None, io=None,
                 write_period=5):
        if location is False or isinstance(location, str):
            self.location = location
        else:
            raise ValueError('unrecognized location ' + repr(location))

        # TODO: when you change formatter or io (and there's data present)
        # make it all look unsaved
        self.formatter = formatter or self.default_formatter
        self.io = io or self.default_io

        self.write_period = write_period
        self.last_write = 0
        self.last_store = -1

        self.metadata = {}

        self.arrays = _PrettyPrintDict()
        if arrays:
            self.action_id_map = self._clean_array_ids(arrays)
            for array in arrays:
                self.add_array(array)

        if self.arrays:
            for array in self.arrays.values():
                array.init_data()

    def sync(self):
        """
        Synchronize this DataSet with the DataServer or storage.

        If this DataSet is on the server, asks the server for changes.
        If not, reads the entire DataSet from disk.

        Returns:
            bool: True if this DataSet is live on the server
        """
        # TODO: sync implies bidirectional... and it could be!
        # we should keep track of last sync timestamp and last modification
        # so we can tell whether this one, the other one, or both copies have
        # changed (and I guess throw an error if both did? Would be cool if we
        # could find a robust and intuitive way to make modifications to the
        # version on the DataServer from the main copy)

        # LOCAL DataSet - no need to sync just use local data
        return False

    def fraction_complete(self):
        """
        Get the fraction of this DataSet which has data in it.

        Returns:
            float: the average of all measured (not setpoint) arrays'
                ``fraction_complete()`` values, independent of the individual
                array sizes. If there are no measured arrays, returns zero.
        """
        array_count, total = 0, 0

        for array in self.arrays.values():
            if not array.is_setpoint:
                array_count += 1
                total += array.fraction_complete()

        return total / (array_count or 1)

    def complete(self, delay=1.5):
        """
        Periodically sync the DataSet and display percent complete status.

        Also, each period, execute functions stored in (class attribute)
        ``self.background_functions``. If a function fails, we log its
        traceback and continue on. If any one function fails twice in
        a row, it gets removed.

        Args:
            delay (float): seconds between iterations. Default 1.5
        """
        log.info(
            f'waiting for DataSet <{self.location}> to complete')

        failing = {key: False for key in self.background_functions}

        completed = False
        while True:
            log.info('DataSet: {:.0f}% complete'.format(
                self.fraction_complete() * 100))

            # first check if we're done
            if self.sync() is False:
                completed = True

            # then even if we *are* done, execute the background functions
            # because we want things like live plotting to get the final data
            for key, fn in list(self.background_functions.items()):
                try:
                    log.debug(f"calling {key}: {repr(fn)}")
                    fn()
                    failing[key] = False
                except Exception:
                    log.info(format_exc())
                    if failing[key]:
                        log.warning(
                            'background function {} failed twice in a row, '
                            'removing it'.format(key))
                        del self.background_functions[key]
                    failing[key] = True

            if completed:
                break

            # but only sleep if we're not already finished
            time.sleep(delay)

        log.info(f'DataSet <{self.location}> is complete')

    def get_changes(self, synced_indices):
        """
        Find changes since the last sync of this DataSet.

        Args:
            synced_indices (dict): ``{array_id: synced_index}`` where
                synced_index is the last flat index which has already
                been synced, for any (usually all) arrays in the DataSet.

        Returns:
            Dict[dict]: keys are ``array_id`` for each array with changes,
                values are dicts as returned by ``DataArray.get_changes``
                and required as kwargs to ``DataArray.apply_changes``.
                Note that not all arrays in ``synced_indices`` need be
                present in the return, only those with changes.
        """
        changes = {}

        for array_id, synced_index in synced_indices.items():
            array_changes = self.arrays[array_id].get_changes(synced_index)
            if array_changes:
                changes[array_id] = array_changes

        return changes

    def add_array(self, data_array):
        """
        Add one DataArray to this DataSet, and mark it as part of this DataSet.

        Note: DO NOT just set ``data_set.arrays[id] = data_array``, because
        this will not check if we are overwriting another array, nor set the
        reference back to this DataSet, nor that the ``array_id`` in the array
        matches how you're storing it here.

        Args:
            data_array (DataArray): the new array to add

        Raises:
            ValueError: if there is already an array with this id here.
        """
        # TODO: mask self.arrays so you *can't* set it directly?

        if data_array.array_id in self.arrays:
            raise ValueError('array_id {} already exists in this '
                             'DataSet'.format(data_array.array_id))
        self.arrays[data_array.array_id] = data_array

        # back-reference to the DataSet
        data_array.data_set = self

    def remove_array(self, array_id):
        """ Remove an array from a dataset

        Throws an exception when the array specified is refereced by other
        arrays in the dataset.

        Args:
            array_id (str): array_id of array to be removed
        """
        for a in self.arrays:
            sa = self.arrays[a].set_arrays
            if array_id in [a.array_id for a in sa]:
                raise Exception(
                    'cannot remove array %s as it is referenced by a' % array_id)
        _ = self.arrays.pop(array_id)
        self.action_id_map = self._clean_array_ids(self.arrays.values())

    def _clean_array_ids(self, arrays):
        """
        replace action_indices tuple with compact string array_ids
        stripping off as much extraneous info as possible
        """
        action_indices = [array.action_indices for array in arrays]
        for array in arrays:
            name = array.full_name
            if array.is_setpoint and name and not name.endswith('_set'):
                name += '_set'

            array.array_id = name
        array_ids = {array.array_id for array in arrays}
        for name in array_ids:
            param_arrays = [array for array in arrays
                            if array.array_id == name]
            self._clean_param_ids(param_arrays, name)

        array_ids = [array.array_id for array in arrays]

        return dict(zip(action_indices, array_ids))

    def _clean_param_ids(self, arrays, name):
        # strip off as many leading equal indices as possible
        # and append the rest to the back of the name with underscores
        param_action_indices = [list(array.action_indices) for array in arrays]
        while all(len(ai) for ai in param_action_indices):
            if len({ai[0] for ai in param_action_indices}) == 1:
                for ai in param_action_indices:
                    ai[:1] = []
            else:
                break
        for array, ai in zip(arrays, param_action_indices):
            array.array_id = name + ''.join('_' + str(i) for i in ai)

    def store(self, loop_indices, ids_values):
        """
        Insert data into one or more of our DataArrays.

        Args:
            loop_indices (tuple): the indices within whatever loops we are
                inside. May have fewer dimensions than some of the arrays
                we are inserting into, if the corresponding value makes up
                the remaining dimensionality.
            values (Dict[Union[float, Sequence]]): a dict whose keys are
                array_ids, and values are single numbers or entire slices
                to insert into that array.
         """
        for array_id, value in ids_values.items():
            self.arrays[array_id][loop_indices] = value
        self.last_store = time.time()
        if (self.write_period is not None and
                time.time() > self.last_write + self.write_period):
            log.debug('Attempting to write')
            self.write()
            self.last_write = time.time()
        # The below could be useful but as it writes at every single
        # step of the loop its too verbose even at debug
        # else:
        #     log.debug('.store method: This is not the right time to write')

    def default_parameter_name(self, paramname: Optional[str] = None) -> Optional[str]:
        """Return name of default parameter for plotting

        The default parameter is determined by looking into
        metdata['default_parameter_name'].  If this variable is not present,
        then the closest match to the argument paramname is tried.

        Args:
            paramname: Name to match to parameter name

        Returns:
            Name of the default parameter, or None if no parameter is found
        """
        arraynames = self.arrays.keys()

        # overrule parameter name from the metadata
        if self.metadata.get('default_parameter_name', False):
            paramname = self.metadata['default_parameter_name']

        # try to return the exact name
        if paramname in arraynames:
            return paramname

        # try find something similar
        if paramname is not None:
            vv = [v for v in arraynames if v.endswith(paramname)]
            if len(vv) > 0:
                return vv[0]
            vv = [v for v in arraynames if v.startswith(paramname)]
            if len(vv) > 0:
                return vv[0]

        # try to get the first non-setpoint array
        vv = [v for v in arraynames if not self.arrays[v].is_setpoint]
        if (len(vv) > 0):
            return sorted(vv)[0]

        # fallback: any array found
        try:
            name = sorted(list(arraynames))[0]
            return name
        except IndexError:
            pass
        return None

    def default_parameter_array(self, paramname='amplitude'):
        """ Return default parameter array

        Args:
            paramname (str): Name to match to parameter name.
                 Defaults to 'amplitude'

        Returns:
            DataArray: array corresponding to the default parameter

        See also:
            default_parameter_name

        """
        paramname = self.default_parameter_name(paramname=paramname)
        return getattr(self, paramname, None)

    def read(self):
        """Read the whole DataSet from storage, overwriting the local data."""
        if self.location is False:
            return
        self.formatter.read(self)

    def read_metadata(self):
        """Read the metadata from storage, overwriting the local data."""
        if self.location is False:
            return
        self.formatter.read_metadata(self)

    def write(self, write_metadata=False, only_complete=True, filename=None):
        """
        Writes updates to the DataSet to storage.
        N.B. it is recommended to call data_set.finalize() when a DataSet is
        no longer expected to change to ensure files get closed

        Args:
            write_metadata (bool): write the metadata to disk
            only_complete (bool): passed on to the match_save_range inside
                self.formatter.write. Used to ensure that all new data gets
                saved even when some columns are strange.
            filename (Optional[str]): The filename (minus extension) to use.
                The file gets saved in the usual location.
        """
        if self.location is False:
            return

        # Only the gnuplot formatter has a "filename" kwarg
        if isinstance(self.formatter, GNUPlotFormat):
            self.formatter.write(self,
                                 self.io,
                                 self.location,
                                 write_metadata=write_metadata,
                                 only_complete=only_complete,
                                 filename=filename)
        else:
            self.formatter.write(self,
                                 self.io,
                                 self.location,
                                 write_metadata=write_metadata,
                                 only_complete=only_complete)

    def write_copy(self, path=None, io_manager=None, location=None):
        """
        Write a new complete copy of this DataSet to storage.

        Args:
            path (Optional[str]): An absolute path on this system to write to.
                If you specify this, you may not include either ``io_manager``
                or ``location``.

            io_manager (Optional[io_manager]): A new ``io_manager`` to use with
                either the ``DataSet``'s same or a new ``location``.

            location (Optional[str]): A new ``location`` to write to, using
                either this ``DataSet``'s same or a new ``io_manager``.
        """
        if io_manager is not None or location is not None:
            if path is not None:
                raise TypeError('If you provide io_manager or location '
                                'to write_copy, you may not provide path.')
            if io_manager is None:
                io_manager = self.io
            elif location is None:
                location = self.location
        elif path is not None:
            io_manager = DiskIO(None)
            location = path
        else:
            raise TypeError('You must provide at least one argument '
                            'to write_copy')

        if location is False:
            raise ValueError('write_copy needs a location, not False')

        lsi_cache = {}
        mr_cache = {}
        for array_id, array in self.arrays.items():
            lsi_cache[array_id] = array.last_saved_index
            mr_cache[array_id] = array.modified_range
            # array.clear_save() is not enough, we _need_ to set modified_range
            # TODO - identify *when* clear_save is not enough, and fix it
            # so we *can* use it. That said, maybe we will *still* want to
            # use the full array here no matter what, or strip trailing NaNs
            # separately, either here or in formatter.write?
            array.last_saved_index = None
            array.modified_range = (0, array.ndarray.size - 1)

        try:
            self.formatter.write(self, io_manager, location, force_write=True)
            self.snapshot()
            self.formatter.write_metadata(self, io_manager, location,
                                          read_first=False)
        finally:
            for array_id, array in self.arrays.items():
                array.last_saved_index = lsi_cache[array_id]
                array.modified_range = mr_cache[array_id]

    def add_metadata(self, new_metadata):
        """
        Update DataSet.metadata with additional data.

        Args:
            new_metadata (dict): new data to be deep updated into
                the existing metadata
        """
        deep_update(self.metadata, new_metadata)

    def save_metadata(self):
        """Evaluate and save the DataSet's metadata."""
        if self.location is not False:
            self.snapshot()
            self.formatter.write_metadata(self, self.io, self.location)

    def finalize(self, filename=None, write_metadata=True):
        """
        Mark the DataSet complete and write any remaining modifications.

        Also closes the data file(s), if the ``Formatter`` we're using
        supports that.

        Args:
            filename (Optional[str]): The file name (minus extension) to
                write to. The location of the file is the usual one.
            write_metadata (bool): Whether to save a snapshot. For e.g. dumping
                raw data inside a loop, a snapshot is not wanted.
        """
        log.debug('Finalising the DataSet. Writing.')
        # write all new data, not only (to?) complete columns
        self.write(only_complete=False, filename=filename)

        if hasattr(self.formatter, 'close_file'):
            self.formatter.close_file(self)

        if write_metadata:
            self.save_metadata()

    def snapshot(self, update=False):
        """JSON state of the DataSet."""
        array_snaps = {}
        for array_id, array in self.arrays.items():
            array_snaps[array_id] = array.snapshot(update=update)

        self.metadata.update({
            '__class__': full_class(self),
            'location': self.location,
            'arrays': array_snaps,
            'formatter': full_class(self.formatter),
            'io': repr(self.io)
        })
        return deepcopy(self.metadata)

    def get_array_metadata(self, array_id):
        """
        Get the metadata for a single contained DataArray.

        Args:
            array_id (str): the array to get metadata for.

        Returns:
            dict: metadata for this array.
        """
        try:
            return self.metadata['arrays'][array_id]
        except (AttributeError, KeyError):
            return None

    def __repr__(self):
        """Rich information about the DataSet and contained arrays."""
        out = type(self).__name__ + ':'

        attrs = [['location', repr(self.location)]]
        attr_template = '\n   {:8} = {}'
        for var, val in attrs:
            out += attr_template.format(var, val)

        arr_info = [['<Type>', '<array_id>', '<array.name>', '<array.shape>']]

        if hasattr(self, 'action_id_map'):
            id_items = [
                item for index, item in sorted(self.action_id_map.items())]
        else:
            id_items = self.arrays.keys()

        for array_id in id_items:
            array = self.arrays[array_id]
            setp = 'Setpoint' if array.is_setpoint else 'Measured'
            name = array.name or 'None'
            array_id = array_id or 'None'
            arr_info.append([setp, array_id, name, repr(array.shape)])

        column_lengths = [max(len(row[i]) for row in arr_info)
                          for i in range(len(arr_info[0]))]
        out_template = ('\n   '
                        '{info[0]:{lens[0]}} | {info[1]:{lens[1]}} | '
                        '{info[2]:{lens[2]}} | {info[3]}')

        for arr_info_i in arr_info:
            out += out_template.format(info=arr_info_i, lens=column_lengths)

        return out

    def to_xarray(self) -> xr.Dataset:
        """ Convert the dataset to an xarray Dataset """
        return qcodes_dataset_to_xarray_dataset(self)

    @classmethod
    def from_xarray(cls, xarray_dataset: xr.Dataset) -> 'DataSet':
        """ Convert the dataset to an xarray DataSet """
        return xarray_dataset_to_qcodes_dataset(xarray_dataset)


class _PrettyPrintDict(Dict[Any, Any]):
    """
    simple wrapper for a dict to repr its items on separate lines
    with a bit of indentation
    """

    def __repr__(self):
        body = "\n  ".join(
            repr(k) + ": " + self._indent(repr(v)) for k, v in self.items()
        )
        return "{\n  " + body + "\n}"

    def _indent(self, s):
        lines = s.split('\n')
        return '\n    '.join(lines)


def dataset_to_xarray_dictionary(
    data_set: DataSet, include_metadata: bool = True
) -> Dict[str, Any]:
    """Convert QcodesDataSet to dictionary.

    Args:
        data_set: The data to convert.
        include_data: If True then include the ndarray field.
        include_metadata: If True then include the metadata.

    Returns:
        Dictionary containing the serialized data.
    """
    data_dictionary: Dict[str, Any] = {
        "dims": {},
        "attrs": {},
        "coords": {},
        "data_vars": {},
    }

    pa = data_set.default_parameter_array()
    dimensions = [(a.array_id, a.size) for a in pa.set_arrays]
    data_dictionary["dims"] = dict(dimensions)

    for array_id in [item[0] for item in dimensions]:
        data_array = data_set.arrays[array_id]
        data_dictionary["coords"][array_id] = data_array_to_xarray_dictionary(
            data_array
        )

    for array_id, data_array in data_set.arrays.items():
        if not data_array.is_setpoint:
            data_dictionary["data_vars"][array_id] = data_array_to_xarray_dictionary(
                data_array
            )

    if include_metadata:
        data_dictionary["attrs"]["metadata"] = data_set.metadata
        data_dictionary["attrs"]["qcodes_location"] = data_set.location

    return data_dictionary


def qcodes_dataset_to_xarray_dataset(
    data_set: DataSet,
) -> xr.Dataset:
    """ Convert QCoDeS gridded dataset to xarray dataset """
    xarray_dictionary = dataset_to_xarray_dictionary(data_set)
    xarray_dataset = xr.Dataset.from_dict(xarray_dictionary)
    return xarray_dataset


def xarray_dictionary_to_dataset(
    xarray_dictionary: Dict[str, Any],
) -> DataSet:
    """Convert xarray dictionary to Qcodes DataSet.

    Args:
        xarray_dictionary: data to convert

    Returns:
        QCoDeS dataSet with converted data.
    """
    dataset = new_data()
    dataset.metadata.update(xarray_dictionary["attrs"])

    grid_coords: List[Any] = []
    set_array_names = []
    for array_key, coord_dictionary in xarray_dictionary["coords"].items():
        preset_data = np.array(coord_dictionary["data"])

        tiled_preset_data = np.tile(preset_data, [g.size for g in grid_coords] + [1])
        grid_coords.append(preset_data)

        data_array = xarray_data_array_dictionary_to_data_array(
            array_key, coord_dictionary, True, preset_data=tiled_preset_data
        )
        dataset.add_array(data_array)
        set_array_names.append(array_key)
    for array_key, datavar_dictionary in xarray_dictionary["data_vars"].items():
        set_arrays = tuple(dataset.arrays[name] for name in set_array_names)

        data_array = xarray_data_array_dictionary_to_data_array(
            array_key, datavar_dictionary, False
        )
        data_array.set_arrays = set_arrays
        dataset.add_array(data_array)

    return dataset


def xarray_dataset_to_qcodes_dataset(xarray_data_set: xr.Dataset) -> DataSet:
    """ Convert QCoDeS gridded dataset to xarray dataset """
    xarray_dictionary = xarray_data_set.to_dict()
    qcodes_dataset = xarray_dictionary_to_dataset(xarray_dictionary)

    return qcodes_dataset



import re

import numpy as np
import pytest

import qcodes as qc
from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.measurements import DataSaver

CALLBACK_COUNT = 0
CALLBACK_RUN_ID = None
CALLBACK_SNAPSHOT = None


def callback(result_list, data_set_len, state, run_id, snapshot):
    """
    default_callback example function implemented in the Web UI.
    """
    global CALLBACK_COUNT, CALLBACK_RUN_ID, CALLBACK_SNAPSHOT
    CALLBACK_COUNT += 1
    CALLBACK_RUN_ID = run_id
    CALLBACK_SNAPSHOT = snapshot


def reset_callback_globals():
    global CALLBACK_COUNT, CALLBACK_RUN_ID, CALLBACK_SNAPSHOT
    CALLBACK_COUNT = 0
    CALLBACK_RUN_ID = None
    CALLBACK_SNAPSHOT = None


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_default_callback(bg_writing):
    """
    The Web UI needs to know the results of an experiment with the metadata.
    So a default_callback class variable is set by the Web UI with a callback
    to introspect the data.
    """
    test_set = None
    reset_callback_globals()

    try:
        DataSaver.default_callback = {
            'run_tables_subscription_callback': callback,
            'run_tables_subscription_min_wait': 1,
            'run_tables_subscription_min_count': 2}

        test_set = qc.new_data_set("test-dataset")
        test_set.add_metadata('snapshot', 'reasonable_snapshot')
        DataSaver(dataset=test_set,
                  write_period=0,
                  interdeps=InterDependencies_)
        test_set.mark_started(start_bg_writer=bg_writing)
        test_set.mark_completed()
        assert CALLBACK_SNAPSHOT == 'reasonable_snapshot'
        assert CALLBACK_RUN_ID > 0
        assert CALLBACK_COUNT > 0
    finally:
        DataSaver.default_callback = None
        if test_set is not None:
            test_set.conn.close()


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize("bg_writing", [True, False])
def test_numpy_types(bg_writing):
    """
    Test that we can save numpy types in the data set
    """

    p = ParamSpecBase(name="p", paramtype="numeric")
    test_set = qc.new_data_set("test-dataset")
    test_set.set_interdependencies(InterDependencies_(standalones=(p,)))
    test_set.mark_started(start_bg_writer=bg_writing)

    idps = InterDependencies_(standalones=(p,))

    data_saver = DataSaver(
        dataset=test_set, write_period=0, interdeps=idps)

    dtypes = [np.int8, np.int16, np.int32, np.int64, np.float16, np.float32,
              np.float64]

    for dtype in dtypes:
        data_saver.add_result(("p", dtype(2)))

    data_saver.flush_data_to_database()
    test_set.mark_completed()
    data = test_set.get_parameter_data("p")["p"]["p"]
    expected_data = np.ones(len(dtypes))
    expected_data[:] = 2
    np.testing.assert_array_equal(data, expected_data)


@pytest.mark.usefixtures("experiment")
@pytest.mark.parametrize('numeric_type',
                         [int, float, np.int8, np.int16, np.int32, np.int64,
                          np.float16, np.float32, np.float64])
@pytest.mark.parametrize("bg_writing", [True, False])
def test_saving_numeric_values_as_text(numeric_type, bg_writing):
    """
    Test the saving numeric values into 'text' parameter raises an exception
    """
    p = ParamSpecBase("p", "text")

    test_set = qc.new_data_set("test-dataset")
    test_set.set_interdependencies(InterDependencies_(standalones=(p,)))
    test_set.mark_started(start_bg_writer=bg_writing)

    idps = InterDependencies_(standalones=(p,))

    data_saver = DataSaver(
        dataset=test_set, write_period=0, interdeps=idps)

    try:
        value = numeric_type(2)

        gottype = np.array(value).dtype

        msg = re.escape(f'Parameter {p.name} is of type '
                        f'"{p.type}", but got a result of '
                        f'type {gottype} ({value}).')
        with pytest.raises(ValueError, match=msg):
            data_saver.add_result((p.name, value))
    finally:
        data_saver.dataset.mark_completed()
        data_saver.dataset.conn.close()


@pytest.mark.usefixtures("experiment")
def test_duplicated_parameter_raises():
    """
    Test that passing same parameter multiple times to ``add_result`` raises an exception
    """
    p = ParamSpecBase("p", "text")

    test_set = qc.new_data_set("test-dataset")
    test_set.set_interdependencies(InterDependencies_(standalones=(p,)))
    test_set.mark_started()

    idps = InterDependencies_(standalones=(p,))

    data_saver = DataSaver(dataset=test_set, write_period=0, interdeps=idps)

    try:
        msg = re.escape(
            "Not all parameter names are unique. Got multiple values for ['p']"
        )
        with pytest.raises(ValueError, match=msg):
            data_saver.add_result((p.name, 1), (p.name, 1))
    finally:
        data_saver.dataset.mark_completed()
        data_saver.dataset.conn.close()



from unittest.mock import MagicMock

import pytest
from pyvisa import VisaIOError

from qcodes.instrument_drivers.Keysight.keysightb1500 import constants
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1500_base \
    import KeysightB1500
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1511B import \
    B1511B
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1517A import \
    B1517A
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1520A import \
    B1520A
from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1530A import \
    B1530A
from qcodes.instrument_drivers.Keysight.keysightb1500.constants import \
    SlotNr, ChNr, CALResponse

# pylint: disable=redefined-outer-name


@pytest.fixture
def b1500(request):
    request.addfinalizer(KeysightB1500.close_all)

    try:
        resource_name = 'insert_Keysight_B2200_VISA_resource_name_here'
        instance = KeysightB1500('SPA',
                                 address=resource_name)
    except (ValueError, VisaIOError):
        # Either there is no VISA lib installed or there was no real
        # instrument found at the specified address => use simulated instrument
        import qcodes.instrument.sims as sims
        path_to_yaml = sims.__file__.replace('__init__.py',
                                             'keysight_b1500.yaml')

        instance = KeysightB1500('SPA',
                                 address='GPIB::1::INSTR',
                                 visalib=path_to_yaml + '@sim'
                                 )

    instance.get_status()
    instance.reset()

    yield instance


def test_make_module_from_model_name():
    mainframe = MagicMock()

    with pytest.raises(NotImplementedError):
        KeysightB1500.from_model_name(model='unsupported_module', slot_nr=0,
                                      parent=mainframe, name='dummy')

    smu = KeysightB1500.from_model_name(model='B1517A', slot_nr=1,
                                        parent=mainframe, name='dummy1')

    assert isinstance(smu, B1517A)

    cmu = KeysightB1500.from_model_name(model='B1520A', slot_nr=2,
                                        parent=mainframe)

    assert isinstance(cmu, B1520A)

    wgfmu = KeysightB1500.from_model_name(model='B1530A', slot_nr=3,
                                          parent=mainframe)

    assert isinstance(wgfmu, B1530A)

    smu = KeysightB1500.from_model_name(model='B1511B', slot_nr=4,
                                        parent=mainframe, name='dummy2')

    assert isinstance(smu, B1511B)


def test_init(b1500):
    assert hasattr(b1500, 'smu1')
    assert hasattr(b1500, 'smu2')
    assert hasattr(b1500, 'cmu1')
    assert hasattr(b1500, 'wgfmu1')


def test_snapshot_does_not_raise_warnings(b1500):
    with pytest.warns(None) as warnings_record:
        b1500.snapshot(update=True)
    assert len(warnings_record) == 0, warnings_record


def test_submodule_access_by_class(b1500):
    assert b1500.smu1 in b1500.by_kind['SMU']
    assert b1500.smu2 in b1500.by_kind['SMU']
    assert b1500.cmu1 in b1500.by_kind['CMU']
    assert b1500.wgfmu1 in b1500.by_kind['WGFMU']


def test_submodule_access_by_slot(b1500):
    assert b1500.smu1 is b1500.by_slot[SlotNr.SLOT01]
    assert b1500.smu2 is b1500.by_slot[SlotNr.SLOT02]
    assert b1500.cmu1 is b1500.by_slot[3]
    assert b1500.wgfmu1 is b1500.by_slot[6]


def test_submodule_access_by_channel(b1500):
    assert b1500.smu1 is b1500.by_channel[ChNr.SLOT_01_CH1]
    assert b1500.smu2 is b1500.by_channel[ChNr.SLOT_02_CH1]
    assert b1500.cmu1 is b1500.by_channel[ChNr.SLOT_03_CH1]
    assert b1500.wgfmu1 is b1500.by_channel[ChNr.SLOT_06_CH1]
    assert b1500.wgfmu1 is b1500.by_channel[6]
    assert b1500.wgfmu1 is b1500.by_channel[ChNr.SLOT_06_CH2]


def test_enable_multiple_channels(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.enable_channels([1, 2, 3])

    mock_write.assert_called_once_with("CN 1,2,3")


def test_disable_multiple_channels(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.disable_channels([1, 2, 3])

    mock_write.assert_called_once_with("CL 1,2,3")


def test_use_nplc_for_high_speed_adc(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.use_nplc_for_high_speed_adc()
    mock_write.assert_called_once_with("AIT 0,2")

    mock_write.reset_mock()

    b1500.use_nplc_for_high_speed_adc(3)
    mock_write.assert_called_once_with("AIT 0,2,3")


def test_use_nplc_for_high_resolution_adc(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.use_nplc_for_high_resolution_adc()
    mock_write.assert_called_once_with("AIT 1,2")

    mock_write.reset_mock()

    b1500.use_nplc_for_high_resolution_adc(8)
    mock_write.assert_called_once_with("AIT 1,2,8")


def test_autozero_enabled(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    assert b1500.autozero_enabled() is False

    b1500.autozero_enabled(True)
    mock_write.assert_called_once_with("AZ 1")
    assert b1500.autozero_enabled() is True

    mock_write.reset_mock()

    b1500.autozero_enabled(False)
    mock_write.assert_called_once_with("AZ 0")
    assert b1500.autozero_enabled() is False


def test_use_manual_mode_for_high_speed_adc(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.use_manual_mode_for_high_speed_adc()
    mock_write.assert_called_once_with("AIT 0,1")

    mock_write.reset_mock()

    b1500.use_manual_mode_for_high_speed_adc(n=1)
    mock_write.assert_called_once_with("AIT 0,1,1")

    mock_write.reset_mock()

    b1500.use_manual_mode_for_high_speed_adc(n=8)
    mock_write.assert_called_once_with("AIT 0,1,8")


def test_self_calibration_successful(b1500):
    mock_ask = MagicMock()
    b1500.ask = mock_ask

    mock_ask.return_value = '0'

    response = b1500.self_calibration()

    assert response == CALResponse(0)
    mock_ask.assert_called_once_with('*CAL?')


def test_self_calibration_failed(b1500):
    mock_ask = MagicMock()
    b1500.ask = mock_ask

    expected_response = CALResponse(1) + CALResponse(64)
    mock_ask.return_value = '65'

    response = b1500.self_calibration()

    assert response == expected_response
    mock_ask.assert_called_once_with('*CAL?')


def test_error_message(b1500):
    response = b1500.error_message()
    assert '+0,"No Error."' == response


def test_clear_timer_count(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.clear_timer_count()
    mock_write.assert_called_once_with('TSR')

    mock_write.reset_mock()

    b1500.clear_timer_count(1)
    mock_write.assert_called_once_with('TSR 1')


def test_set_measuremet_mode(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.set_measurement_mode(mode=constants.MM.Mode.SPOT, channels=[1, 2])
    mock_write.assert_called_once_with('MM 1,1,2')


def test_get_measurement_mode(b1500):
    mock_ask = MagicMock()
    b1500.ask = mock_ask

    mock_ask.return_value = 'MM 1,1,2'
    measurement_mode = b1500.get_measurement_mode()
    assert measurement_mode['mode'] == constants.MM.Mode(1)
    assert measurement_mode['channels'] == [1, 2]


def test_get_response_format_and_mode(b1500):
    mock_ask = MagicMock()
    b1500.ask = mock_ask

    mock_ask.return_value = 'FMT 1,1'
    measurement_mode = b1500.get_response_format_and_mode()
    assert measurement_mode['format'] == constants.FMT.Format(1)
    assert measurement_mode['mode'] == constants.FMT.Mode(1)


def test_enable_smu_filters(b1500):
    mock_write = MagicMock()
    b1500.write = mock_write

    b1500.enable_smu_filters(True)
    mock_write.assert_called_once_with('FL 1')

    mock_write.reset_mock()

    b1500.enable_smu_filters(True, [constants.ChNr.SLOT_01_CH1,
                                    constants.ChNr.SLOT_02_CH1,
                                    constants.ChNr.SLOT_03_CH1])
    mock_write.assert_called_once_with('FL 1,1,2,3')

    mock_write.reset_mock()

    b1500.enable_smu_filters(True, [constants.ChNr.SLOT_01_CH2,
                                    constants.ChNr.SLOT_02_CH2,
                                    constants.ChNr.SLOT_03_CH2])
    mock_write.assert_called_once_with('FL 1,102,202,302')


def test_error_message_is_called_after_setting_a_parameter(b1500):
    mock_ask = MagicMock()
    b1500.ask = mock_ask
    mock_ask.return_value = '+0,"No Error."'

    b1500.enable_smu_filters(True)
    mock_ask.assert_called_once_with('ERRX?')

    mock_ask.reset_mock()

    mock_ask.return_value = '+200, "Output channel not enabled"'
    with pytest.raises(Exception) as e_info:
        b1500.enable_smu_filters(True)
    mock_ask.assert_called_once_with('ERRX?')
    error_string = 'While setting this parameter received error: +200, ' \
                  '"Output channel not enabled"'
    assert e_info.value.args[0] == error_string



""" Base class for the channel of an instrument """
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

from ..utils.helpers import full_class
from ..utils.metadata import Metadatable
from ..utils.validators import Validator
from .base import Instrument, InstrumentBase
from .parameter import (
    ArrayParameter,
    Iterator,
    MultiParameter,
    Parameter,
    ParamRawDataType,
)


class InstrumentChannel(InstrumentBase):
    """
    Base class for a channel in an instrument

    Args:
        parent: The instrument to which this channel should be
          attached.

        name: The name of this channel.

    """

    def __init__(self,
                 parent: InstrumentBase,
                 name: str,
                 **kwargs: Any) -> None:
        # need to specify parent before `super().__init__` so that the right
        # `full_name` is available in that scope. `full_name` is used for
        # registering the filter for the log messages. It is composed by
        # iteratively concatenating the full names of the parent instruments
        # scope of `Base`
        self._parent = parent
        super().__init__(name=name, **kwargs)
        # Naming insanity:
        # (see https://github.com/QCoDeS/Qcodes/issues/1140 for a nice table)
        # this has been a confusion about names. don't use name but
        # full_name, or short_name.
        self._name = f"{parent.name}_{str(name)}"

    def __repr__(self) -> str:
        """Custom repr to give parent information"""
        return '<{}: {} of {}: {}>'.format(type(self).__name__,
                                           self.name,
                                           type(self._parent).__name__,
                                           self._parent.name)

    # Pass any commands to read or write from the instrument up to the parent
    def write(self, cmd: str) -> None:
        return self._parent.write(cmd)

    def write_raw(self, cmd: str) -> None:
        return self._parent.write_raw(cmd)

    def ask(self, cmd: str) -> str:
        return self._parent.ask(cmd)

    def ask_raw(self, cmd: str) -> str:
        return self._parent.ask_raw(cmd)

    @property
    def parent(self) -> InstrumentBase:
        return self._parent

    @property
    def root_instrument(self) -> InstrumentBase:
        return self._parent.root_instrument

    @property
    def name_parts(self) -> List[str]:
        name_parts = self._parent.name_parts
        name_parts.append(self.short_name)
        return name_parts


class MultiChannelInstrumentParameter(MultiParameter):
    """
    Parameter to get or set multiple channels simultaneously.

    Will normally be created by a :class:`ChannelList` and not directly by
    anything else.

    Args:
        channels: A list of channels which we can operate on
          simultaneously.
        param_name: Name of the multichannel parameter
    """
    def __init__(self,
                 channels: Sequence[InstrumentChannel],
                 param_name: str,
                 *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._channels = channels
        self._param_name = param_name

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        """
        Return a tuple containing the data from each of the channels in the
        list.
        """
        return tuple(chan.parameters[self._param_name].get() for chan
                     in self._channels)

    def set_raw(self, value: ParamRawDataType) -> None:
        """
        Set all parameters to this value.

        Args:
            value: The value to set to. The type is given by the
                underlying parameter.
        """
        for chan in self._channels:
            getattr(chan, self._param_name).set(value)

    @property
    def full_names(self) -> Tuple[str, ...]:
        """
        Overwrite full_names because the instrument name is already included
        in the name. This happens because the instrument name is included in
        the channel name merged into the parameter name above.
        """

        return self.names


class ChannelList(Metadatable):
    """
    Container for channelized parameters that allows for sweeps over
    all channels, as well as addressing of individual channels.

    Args:
        parent: The instrument to which this channel
            should be attached.

        name: The name of the channel list.

        chan_type: The type of channel contained
            within this list.

        chan_list: An optional iterable of
            channels of type ``chan_type``.  This will create a list and
            immediately lock the :class:`ChannelList`.

        snapshotable: Optionally disables taking of snapshots
            for a given channel list.  This is used when objects
            stored inside a channel list are accessible in multiple
            ways and should not be repeated in an instrument snapshot.

        multichan_paramclass: The class of
            the object to be returned by the ``__getattr__``
            method of :class:`ChannelList`.
            Should be a subclass of :class:`MultiChannelInstrumentParameter`.

    Raises:
        ValueError: If ``chan_type`` is not a subclass of
            :class:`InstrumentChannel`
        ValueError: If ``multichan_paramclass`` is not a subclass of
            :class:`MultiChannelInstrumentParameter` (note that a class is a
            subclass of itself).

    """

    def __init__(self, parent: InstrumentBase,
                 name: str,
                 chan_type: type,
                 chan_list: Optional[Sequence[InstrumentChannel]] = None,
                 snapshotable: bool = True,
                 multichan_paramclass: type = MultiChannelInstrumentParameter):
        super().__init__()

        self._parent = parent
        self._name = name
        if (not isinstance(chan_type, type) or
                not issubclass(chan_type, InstrumentChannel)):
            raise ValueError("Channel Lists can only hold instances of type"
                             " InstrumentChannel")
        if (not isinstance(multichan_paramclass, type) or
                not issubclass(multichan_paramclass,
                               MultiChannelInstrumentParameter)):
            raise ValueError("multichan_paramclass must be a (subclass of) "
                             "MultiChannelInstrumentParameter")

        self._chan_type = chan_type
        self._snapshotable = snapshotable
        self._paramclass = multichan_paramclass

        self._channel_mapping: Dict[str, InstrumentChannel] = {}
        # provide lookup of channels by name
        # If a list of channels is not provided, define a list to store
        # channels. This will eventually become a locked tuple.
        self._channels: Sequence[InstrumentChannel]
        if chan_list is None:
            self._locked = False
            self._channels = []
        else:
            self._locked = True
            self._channels = tuple(chan_list)
            if self._channels is None:
                raise RuntimeError("Empty channel list")
            self._channel_mapping = {channel.short_name: channel
                                     for channel in self._channels}
            if not all(isinstance(chan, chan_type) for chan in self._channels):
                raise TypeError("All items in this channel list must be of "
                                "type {}.".format(chan_type.__name__))

    def __getitem__(self, i: Union[int, slice, Tuple[int, ...]]) -> \
            Union['InstrumentChannel', 'ChannelList']:
        """
        Return either a single channel, or a new :class:`ChannelList`
        containing only the specified channels

        Args:
            i: Either a single channel index or a slice of channels
              to get
        """
        if isinstance(i, slice):
            return ChannelList(self._parent, self._name, self._chan_type,
                               self._channels[i],
                               multichan_paramclass=self._paramclass)
        elif isinstance(i, tuple):
            return ChannelList(self._parent, self._name, self._chan_type,
                               [self._channels[j] for j in i],
                               multichan_paramclass=self._paramclass)
        return self._channels[i]

    def __iter__(self) -> Iterator['InstrumentChannel']:
        return iter(self._channels)

    def __len__(self) -> int:
        return len(self._channels)

    def __repr__(self) -> str:
        return "ChannelList({!r}, {}, {!r})".format(self._parent,
                                                    self._chan_type.__name__,
                                                    self._channels)

    def __add__(self, other: 'ChannelList') -> 'ChannelList':
        """
        Return a new channel list containing the channels from both
        :class:`ChannelList` self and r.

        Both channel lists must hold the same type and have the same parent.

        Args:
            other: Right argument to add.
        """
        if not isinstance(self, ChannelList) or not isinstance(other,
                                                               ChannelList):
            raise TypeError("Can't add objects of type"
                            " {} and {} together".format(type(self).__name__,
                                                         type(other).__name__))
        if self._chan_type != other._chan_type:
            raise TypeError("Both l and r arguments to add must contain "
                            "channels of the same type."
                            " Adding channels of type "
                            "{} and {}.".format(self._chan_type.__name__,
                                                other._chan_type.__name__))
        if self._parent != other._parent:
            raise ValueError("Can only add channels from the same parent "
                             "together.")

        return ChannelList(self._parent, self._name, self._chan_type,
                           list(self._channels) + list(other._channels))

    def append(self, obj: InstrumentChannel) -> None:
        """
        Append a Channel to this list. Requires that the ChannelList is not
        locked and that the channel is of the same type as the ones in the list.

        Args:
            obj: New channel to add to the list.
        """
        if isinstance(self._channels, tuple) or self._locked:
            raise AttributeError("Cannot append to a locked channel list")
        if not isinstance(obj, self._chan_type):
            raise TypeError("All items in a channel list must be of the same "
                            "type. Adding {} to a list of {}"
                            ".".format(type(obj).__name__,
                                       self._chan_type.__name__))
        self._channel_mapping[obj.short_name] = obj
        self._channels = cast(List[InstrumentChannel], self._channels)
        self._channels.append(obj)

    def clear(self) -> None:
        """
        Clear all items from the channel list.
        """
        if self._locked:
            raise AttributeError("Cannot clear a locked channel list")
        # when not locked the _channels seq is a list
        channels = cast(List['InstrumentChannel'], self._channels)
        channels.clear()
        self._channel_mapping.clear()

    def remove(self, obj: InstrumentChannel) -> None:
        """
        Removes obj from channellist if not locked.

        Args:
            obj: Channel to remove from the list.
        """
        if self._locked:
            raise AttributeError("Cannot remove from a locked channel list")
        else:
            self._channels = cast(List[InstrumentChannel], self._channels)
            self._channels.remove(obj)
            self._channel_mapping.pop(obj.short_name)

    def extend(self, objects: Union[Sequence[InstrumentChannel],
                                    'ChannelList']) -> None:
        """
        Insert an iterable of objects into the list of channels.

        Args:
            objects: A list of objects to add into the
              :class:`ChannelList`.
        """
        # objects may be a generator but we need to iterate over it twice
        # below so copy it into a tuple just in case.
        objects_tuple = tuple(objects)
        if self._locked:
            raise AttributeError("Cannot extend a locked channel list")
        if not all(isinstance(obj, self._chan_type) for obj in objects_tuple):
            raise TypeError("All items in a channel list must be of the same "
                            "type.")
        channels = cast(List[InstrumentChannel], self._channels)
        channels.extend(objects_tuple)
        self._channel_mapping.update({
            obj.short_name: obj for obj in objects
        })
        self._channels = channels

    def index(self, obj: InstrumentChannel) -> int:
        """
        Return the index of the given object

        Args:
            obj: The object to find in the channel list.
        """
        return self._channels.index(obj)

    def insert(self, index: int, obj: InstrumentChannel) -> None:
        """
        Insert an object into the channel list at a specific index.

        Args:
            index: Index to insert object.

            obj: Object of type chan_type to insert.
        """
        if isinstance(self._channels, tuple) or self._locked:
            raise AttributeError("Cannot insert into a locked channel list")
        if not isinstance(obj, self._chan_type):
            raise TypeError("All items in a channel list must be of the same "
                            "type. Adding {} to a list of {}"
                            ".".format(type(obj).__name__,
                                       self._chan_type.__name__))
        self._channels = cast(List[InstrumentChannel], self._channels)
        self._channels.insert(index, obj)

    def get_channel_by_name(
        self, *names: str
    ) -> Union[InstrumentChannel, "ChannelList"]:
        """
        Get a channel by name, or a ChannelList if multiple names are given.

        Args:
            *names: channel names
        """
        if len(names) == 0:
            raise Exception("one or more names must be given")
        if len(names) == 1:
            return self._channel_mapping[names[0]]
        selected_channels = tuple(self._channel_mapping[name] for name in names)
        return ChannelList(
            self._parent,
            self._name,
            self._chan_type,
            selected_channels,
            self._snapshotable,
            self._paramclass,
        )

    def get_validator(self) -> 'ChannelListValidator':
        """
        Returns a validator that checks that the returned object is a channel
        in this channel list
        """
        if not self._locked:
            raise AttributeError("Cannot create a validator for an unlocked "
                                 "channel list")
        return ChannelListValidator(self)

    def lock(self) -> None:
        """
        Lock the channel list. Once this is done, the channel list is
        converted to a tuple and any future changes to the list are prevented.
        """
        if self._locked:
            return

        self._channels = tuple(self._channels)
        self._locked = True

    def snapshot_base(self, update: Optional[bool] = True,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        """
        State of the instrument as a JSON-compatible dict (everything that
        the custom JSON encoder class
        :class:`qcodes.utils.helpers.NumpyJSONEncoder` supports).

        Args:
            update: If True, update the state by querying the
                instrument. If None only update if the state is known to be
                invalid. If False, just use the latest values in memory
                and never update.
            params_to_skip_update: List of parameter names that will be skipped
                in update even if update is True. This is useful if you have
                parameters that are slow to update but can be updated in a
                different way (as in the qdac). If you want to skip the
                update of certain parameters in all snapshots, use the
                ``snapshot_get``  attribute of those parameters instead.

        Returns:
            dict: base snapshot
        """
        if self._snapshotable:
            snap = {'channels': {chan.name: chan.snapshot(update=update)
                                     for chan in self._channels},
                    'snapshotable': self._snapshotable,
                    '__class__': full_class(self),
                    }
        else:
            snap = {'snapshotable': self._snapshotable,
                    '__class__': full_class(self),
                    }
        return snap

    def __getattr__(self, name: str) -> Union[MultiChannelInstrumentParameter,
                                              Callable[..., None],
                                              InstrumentChannel]:
        """
        Return a multi-channel function or parameter that we can use to get or
        set all items in a channel list simultaneously.

        Params:
            name: The name of the parameter or function that we want to
            operate on.
        """
        # Check if this is a valid parameter
        if name in self._channels[0].parameters:
            setpoints = None
            setpoint_names = None
            setpoint_labels = None
            setpoint_units = None
            # We need to construct a MultiParameter object to get each of the
            # values our of each parameter in our list, we don't currently try
            # to construct a multiparameter from a list of multi parameters
            if isinstance(self._channels[0].parameters[name], MultiParameter):
                raise NotImplementedError("Slicing is currently not "
                                          "supported for MultiParameters")
            parameters = cast(List[Union[Parameter, ArrayParameter]],
                              [chan.parameters[name] for chan in self._channels])
            names = tuple(f"{chan.name}_{name}"
                          for chan in self._channels)
            labels = tuple(parameter.label
                           for parameter in parameters)
            units = tuple(parameter.unit
                          for parameter in parameters)

            if isinstance(parameters[0], ArrayParameter):
                arrayparameters = cast(List[ArrayParameter], parameters)
                shapes = tuple(parameter.shape for
                               parameter in arrayparameters)
                if arrayparameters[0].setpoints:
                    setpoints = tuple(parameter.setpoints for
                                      parameter in arrayparameters)
                if arrayparameters[0].setpoint_names:
                    setpoint_names = tuple(parameter.setpoint_names for
                                           parameter in arrayparameters)
                if arrayparameters[0].setpoint_labels:
                    setpoint_labels = tuple(
                        parameter.setpoint_labels
                        for parameter in arrayparameters)
                if arrayparameters[0].setpoint_units:
                    setpoint_units = tuple(parameter.setpoint_units
                                           for parameter in arrayparameters)
            else:
                shapes = tuple(() for _ in self._channels)

            param = self._paramclass(
                self._channels,
                param_name=name,
                name=f"Multi_{name}",
                names=names,
                shapes=shapes,
                instrument=self._parent,
                labels=labels,
                units=units,
                setpoints=setpoints,
                setpoint_names=setpoint_names,
                setpoint_units=setpoint_units,
                setpoint_labels=setpoint_labels,
                bind_to_instrument=False,
            )
            return param

        # Check if this is a valid function
        if name in self._channels[0].functions:
            # We want to return a reference to a function that would call the
            # function for each of the channels in turn.
            def multi_func(*args: Any) -> None:
                for chan in self._channels:
                    chan.functions[name](*args)
            return multi_func

        try:
            return self._channel_mapping[name]
        except KeyError:
            pass

        raise AttributeError('\'{}\' object has no attribute \'{}\''
                             ''.format(self.__class__.__name__, name))

    def __dir__(self) -> List[Any]:
        names = list(super().__dir__())
        if self._channels:
            names += list(self._channels[0].parameters.keys())
            names += list(self._channels[0].functions.keys())
            names += [channel.short_name for channel in self._channels]
        return sorted(set(names))

    def print_readable_snapshot(self, update: bool = False,
                                max_chars: int = 80) -> None:
        if self._snapshotable:
            for channel in self._channels:
                channel.print_readable_snapshot(update=update,
                                                max_chars=max_chars)


class ChannelListValidator(Validator[InstrumentChannel]):
    """
    A validator that checks that the returned object is a member of the
    channel list with which the validator was constructed.

    This class will not normally be created directly, but created from a channel
    list using the ``ChannelList.get_validator`` method.

    Args:
        channel_list (ChannelList): the channel list that should be checked
            against. The channel list must be locked and populated before it
            can be used to construct a validator.
    """
    def __init__(self, channel_list: ChannelList) -> None:
        # Save the base parameter list
        if not isinstance(channel_list, ChannelList):
            raise ValueError("channel_list must be a ChannelList "
                             "object containing the "
                             "channels that should be validated")
        if not channel_list._locked:
            raise AttributeError("Channel list must be locked before it can "
                                 "be used to create a validator")
        self._channel_list = channel_list

    def validate(self, value: InstrumentChannel, context: str = '') -> None:
        """
        Checks to see that value is a member of the channel list referenced by
        this validator

        Args:
            value: the value to be checked against the
                reference channel list.
            context: the context of the call, used as part of the exception
                raised.
        """
        if value not in self._channel_list:
            raise ValueError(
                '{} is not part of the expected channel list; {}'.format(
                    repr(value), context))


class AutoLoadableInstrumentChannel(InstrumentChannel):
    """
    This subclass provides extensions to auto-load channels
    from instruments and adds methods to create and delete
    channels when possible. Please note that `channel` in this
    context does not necessarily mean a physical instrument channel,
    but rather an instrument sub-module. For some instruments,
    these sub-modules can be created and deleted at will.
    """

    @classmethod
    def load_from_instrument(
            cls, parent: Instrument,
            channel_list: Optional['AutoLoadableChannelList'] = None,
            **kwargs: Any
    ) -> List['AutoLoadableInstrumentChannel']:
        """
        Load channels that already exist on the instrument

        Args:
            parent: The instrument through which the instrument
                channel is accessible
            channel_list: The channel list this
                channel is a part of
            **kwargs: Keyword arguments needed to create the channels

        Returns:
            List of instrument channel instances created for channels
            that already exist on the instrument
        """

        obj_list = []
        for new_kwargs in cls._discover_from_instrument(parent, **kwargs):
            obj = cls(
                parent, existence=True, channel_list=channel_list,
                **new_kwargs
            )
            obj_list.append(obj)

        return obj_list

    @classmethod
    def _discover_from_instrument(
            cls, parent: Instrument, **kwargs: Any) -> List[Dict[Any, Any]]:
        """
        Discover channels on the instrument and return a list kwargs to create
        these channels in memory

        Args:
            parent: The instrument through which the instrument
                channel is accessible
            **kwargs: Keyword arguments needed to discover the channels

        Returns:
              List of keyword arguments for channel instance initialization
              for each channel that already exists on the physical instrument
        """
        raise NotImplementedError(
            "Please subclass and implement this method in the subclass")

    @classmethod
    def new_instance(
            cls, parent: Instrument, create_on_instrument: bool = True,
            channel_list: Optional['AutoLoadableChannelList'] = None,
            **kwargs: Any
    ) -> 'AutoLoadableInstrumentChannel':
        """
        Create a new instance of the channel on the instrument: This involves
        finding initialization arguments which will create a channel with a
        unique name and create the channel on the instrument.

        Args:
            parent: The instrument through which the instrument
                channel is accessible
            create_on_instrument: When True, the channel is immediately
                created on the instrument
            channel_list: The channel list this
                channel is going to belong to
            **kwargs: Keyword arguments needed to create a new instance.
        """
        new_kwargs = cls._get_new_instance_kwargs(parent=parent, **kwargs)

        try:
            new_instance = cls(parent, channel_list=channel_list, **new_kwargs)
        except TypeError as err:
            # The 'new_kwargs' dict is malformed. Investigate more precisely
            # why and give the user a more helpful hint how this can be
            # solved.
            if "name" not in new_kwargs:
                raise TypeError(
                    "A 'name' argument should be supplied by the "
                    "'_get_new_instance_kwargs' method"
                ) from err
            if "parent" in new_kwargs:
                raise TypeError(
                    "A 'parent' argument should *not* be supplied by the "
                    "'_get_new_instance_kwargs' method"
                ) from err
            # Something else has gone wrong. Probably, not all mandatory keyword
            # arguments are supplied
            raise TypeError(
                "Probably, the '_get_new_instance_kwargs' method does not "
                "return all of the required keyword arguments") from err

        if create_on_instrument:
            new_instance.create()

        return new_instance

    @classmethod
    def _get_new_instance_kwargs(cls, parent: Optional[Instrument] = None,
                                 **kwargs: Any) -> Dict[Any, Any]:
        """
        Returns a dictionary which is used as keyword args when instantiating a
        channel

        Args:
            parent: The instrument the new channel will belong to. Not all
                instruments need this so it is an optional argument
            **kwargs: Additional arguments which are needed to
                instantiate a channel can be given directly by the calling
                function.

        Returns:
            A keyword argument dictionary with at least a ``name`` key which is
            unique on the instrument. The parent instrument is passed as an
            argument in this function so we can query if the generated name is
            indeed unique.

        Notes:
            The init arguments ``parent`` and ``channel_list`` are automatically
            added by the ``new_instance`` method and should not be added in the
            kwarg dictionary returned here. Additionally, the argument
            ``existence`` either needs to be omitted or be False.
        """
        raise NotImplementedError(
            "Please subclass and implement this method in the subclass")

    def __init__(
            self,
            parent: Union[Instrument, 'InstrumentChannel'],
            name: str,
            exists_on_instrument: bool = False,
            channel_list: Optional['AutoLoadableChannelList'] = None,
            **kwargs: Any):
        """
        Instantiate a channel object. Note that this is not the same as actually
        creating the channel on the instrument. Parameters defined on this
        channels will not be able to query/write to the instrument until it
        has been created on the instrument

        Args:
            parent: The instrument through which the instrument
                channel is accessible
            name: channel name
            exists_on_instrument: True if the channel exists on the instrument
            channel_list: Reference to the list that this channel is a member
                of; this is used when deleting the channel so that it can remove
                itself from the list
        """
        super().__init__(parent, name=name, **kwargs)
        self._exists_on_instrument = exists_on_instrument
        self._channel_list = channel_list

    def create(self) -> None:
        """Create the channel on the instrument"""
        if self._exists_on_instrument:
            raise RuntimeError("Channel already exists on instrument")

        self._create()
        self._exists_on_instrument = True

    def _create(self) -> None:
        """
        (SCPI) commands needed to create the channel. Note that we
        need to use ``self.root_instrument.write`` to send commands,
        because ``self.write`` will cause ``_assert_existence`` to raise a
        runtime error.
        """
        raise NotImplementedError("Please subclass")

    def remove(self) -> None:
        """
        Delete the channel from the instrument and remove from channel list
        """
        self._assert_existence()
        self._remove()
        if self._channel_list is not None and self in self._channel_list:
            self._channel_list.remove(self)
        self._exists_on_instrument = False

    def _remove(self) -> None:
        """
        (SCPI) commands needed to delete the channel from the instrument
        """
        raise NotImplementedError("Please subclass")

    def _assert_existence(self) -> None:
        if not self._exists_on_instrument:
            raise RuntimeError(
                "Object does not exist (anymore) on the instrument")

    def write(self, cmd: str) -> None:
        """
        Write to the instrument only if the channel is present on the instrument
        """
        self._assert_existence()
        return super().write(cmd)

    def ask(self, cmd: str) -> str:
        """
        Ask the instrument only if the channel is present on the instrument
        """
        self._assert_existence()
        return super().ask(cmd)

    @property
    def exists_on_instrument(self) -> bool:
        return self._exists_on_instrument


class AutoLoadableChannelList(ChannelList):
    """
    Extends the QCoDeS :class:`ChannelList` class to add the following features:
    - Automatically create channel objects on initialization
    - Make a ``add`` method to create channel objects

    Args:
        parent: the instrument to which this channel
            should be attached

        name: the name of the channel list

        chan_type: the type of channel contained
            within this list

        chan_list: An optional iterable of
            channels of type chan_type.  This will create a list and
            immediately lock the :class:`ChannelList`.

        snapshotable: Optionally disables taking of snapshots
            for a given channel list.  This is used when objects
            stored inside a channel list are accessible in multiple
            ways and should not be repeated in an instrument snapshot.

        multichan_paramclass: The class of
            the object to be returned by the
            :class:`ChannelList` ``__getattr__`` method.
            Should be a subclass of :class:`MultiChannelInstrumentParameter`.

        **kwargs: Keyword arguments to be passed to the ``load_from_instrument``
            method of the channel class. Note that the kwargs are *NOT* passed
            to the ``__init__`` of the super class.

    Raises:
        ValueError: If :class:`chan_type` is not a subclass of
            :class:`InstrumentChannel`
        ValueError: If ``multichan_paramclass`` is not a subclass of
            :class:`MultiChannelInstrumentParameter` (note that a class is a
            subclass of itself).
    """
    def __init__(
            self,
            parent: Instrument,
            name: str,
            chan_type: type,
            chan_list: Optional[Sequence['AutoLoadableInstrumentChannel']] = None,
            snapshotable: bool = True,
            multichan_paramclass: type = MultiChannelInstrumentParameter,
            **kwargs: Any
    ) -> None:

        super().__init__(
            parent, name, chan_type, chan_list, snapshotable,
            multichan_paramclass
        )
        new_channels = self._chan_type.load_from_instrument(  # type: ignore[attr-defined]
            self._parent, channel_list=self, **kwargs)

        self.extend(new_channels)

    def add(self, **kwargs: Any) -> 'AutoLoadableInstrumentChannel':
        """
        Add a channel to the list

        Args:
            kwargs: Keyword arguments passed to the ``new_instance`` method of
                the channel class

        Returns:
            Newly created instance of the channel class
        """
        new_channel = self._chan_type.new_instance(  # type: ignore[attr-defined]
            self._parent,
            create_on_instrument=True,
            channel_list=self,
            **kwargs
        )

        self.append(new_channel)
        return new_channel
